"""Low-level implementation of resource download."""

from __future__ import annotations

import dataclasses
import hashlib
import io
import pathlib
import typing

import requests

from . import constants, exception, task, utilities


@dataclasses.dataclass
class Progress:
    """Message that reports download progress."""

    path_id: pathlib.PurePosixPath
    """Path ID of the associated resource
    """

    initial_bytes: int
    """Number of bytes of the remote resource that were already downloaded when the current download began.
    """

    current_bytes: int
    """Number of bytes of the remote resource that have been downloaded so far.
    """

    final_bytes: int
    """Total number of bytes of the remote resource.
    """

    complete: bool
    """Whether this resource has been completely downloaded.
    """


@dataclasses.dataclass(frozen=True)
class Server:
    """Represents a remote server."""

    url: str
    """The server's base URL.

    Resources URL are calculated by appending the file's path ID to the server URL. A slash is inserted before the path ID if the server's URL does not end with one.
    """

    timeout: float
    """Timeout in seconds for requests to this server.
    """

    def path_id_to_url(self, path_id: pathlib.PurePosixPath) -> str:
        """Calculates a resource URL from its path ID.

        Args:
            path_id (pathlib.PurePosixPath): The resource's path ID, including the dataset name.

        Returns:
            str: The resource's remote URL.
        """
        if len(path_id.parts) == 1:
            return self.url
        return "{}{}{}".format(
            self.url,
            "" if self.url.endswith("/") else "/",
            "/".join(path_id.parts[1:]),
        )


@dataclasses.dataclass(frozen=True)
class NullServer(Server):
    """A placeholder server that raises an exception when used.

    Some functions and classes require a server to download resources that are no available locally.
    If the resources are known to be local, this server can be used to detect download attempts.
    """

    def __init__(self):
        super().__init__(url="", timeout=0.0)

    def path_id_to_url(self, path_id: pathlib.PurePosixPath):
        raise NotImplementedError()


class Download(task.Task):
    """Retrieves data from a remote server.

    This is an abstract task that calls its methods (lifecycle callbacks) as follows:

    - :py:meth:`on_begin` is called before contacting the server.
      This function can be used to create write resources and must return an offset in bytes.
      Download resumes from that offset if it is non-zero.
      If the offset is negative, the task assumes that the download is complete and it calls :py:meth:`on_end` immediately.
    - :py:meth:`on_range_failed` is called if :py:meth:`on_begin` returned a non-zero offset and the server
      rejects the range request (HTTP 206). It can be used to clean up 'append' resources and replace them
      with 'write' resources. The actual download starts after :py:meth:`on_range_failed` as if :py:meth:`on_begin` returned 0.
    - :py:meth:`on_response_ready` is called when the response is ready for iteration.
      The subclass must call :py:meth:`requests.Response.close` after reading the response (and probably :py:meth:`on_end`).

    This lifecycle allows users to yield on response chunks (see :py:meth:`undr.path.File._chunks` for an example).

    Args:
        path_id (pathlib.PurePosixPath): The resource's unique path id.
        suffix (typing.Optional[str]): Added to the file name while it is being downloaded.
        server (Server): The remote server.
        stream (bool): Whether to download the file in chunks (slightly slower for small files, reduces memory usage for large files).
    """

    def __init__(
        self,
        path_id: pathlib.PurePosixPath,
        suffix: typing.Optional[str],
        server: Server,
        stream: bool,
    ):
        self.path_id = path_id
        self.suffix = suffix
        self.server = server
        self.stream = stream

    def url(self) -> str:
        """Returns the file's remote URL.

        Returns:
            str: File URL on the server.
        """
        return self.server.path_id_to_url(
            self.path_id
            if self.suffix is None
            else utilities.posix_path_with_suffix(self.path_id, self.suffix)
        )

    def run(self, session: requests.Session, manager: task.Manager):
        skip = self.on_begin(manager=manager)
        if skip < 0:
            self.on_end(manager=manager)
        else:
            response: typing.Optional[requests.Response] = None
            if skip > 0:
                response = session.get(
                    self.url(),
                    timeout=self.server.timeout,
                    stream=self.stream,
                    headers={"Range": f"bytes={skip}-"},
                )
                if response.status_code != 206:
                    self.on_range_failed(manager=manager)
                    response = None
            if response is None:
                response = session.get(
                    self.url(),
                    timeout=self.server.timeout,
                    stream=self.stream,
                )
            response.raise_for_status()
            self.on_response_ready(response=response, manager=manager)

    def on_begin(self, manager: task.Manager) -> int:
        """Called before contacting the server.

        This function must return an offset in bytes.

        - 0 indicates that the file is not downloaded yet.
        - Positive values indicate the number of bytes already downloaded.
        - Negative values indicate that the download is already complete and must be skipped.

        Args:
            manager (task.Manager): The task manager for reporting updates.

        Returns:
            int: Number of bytes already downloaded.
        """
        raise NotImplementedError()

    def on_range_failed(self, manager: task.Manager) -> None:
        """Called if the HTTP range call fails.

        The HTTP range request asks the serve to resumes download at
        a given byte offset. It used when :py:meth:`on_begin` returns a non-zero value.
        Range is not always supported by the server. This function should reset counters
        and ready the local file system for a standard (full) download.

        Args:
            manager (task.Manager): The task manager for reporting updates.
        """
        raise NotImplementedError()

    def on_response_ready(
        self, response: requests.Response, manager: task.Manager
    ) -> None:
        """Called when the HTTP response object is ready.

        The reponse object can be used to download the remote file.

        Args:
            response (requests.Response): HTTP response object.
            manager (task.Manager): The task manager for reporting updates.
        """
        raise NotImplementedError()

    def on_end(self, manager: task.Manager) -> None:
        """Called when the download task completes.

        This function is called automatically if the byte offset
        returned by :py:meth:`on_begin` is nagative.
        Implementations should call it after consuming the response in :py:meth:`on_response_ready`.

        Args:
            manager (task.Manager): The task manager for reporting updates.
        """
        raise NotImplementedError()


class DownloadFile(Download):
    """Retrieves data from a remote server and saves it to a file."""

    def __init__(
        self,
        path_root: pathlib.Path,
        path_id: pathlib.PurePosixPath,
        suffix: typing.Optional[str],
        server: Server,
        force: bool,
        expected_size: typing.Optional[int],
        expected_hash: typing.Optional[str],
    ):
        super().__init__(
            path_id=path_id,
            suffix=suffix,
            server=server,
            stream=expected_size is None
            or expected_size >= constants.CHUNK_SIZE * constants.STREAM_CHUNK_THRESHOLD,
        )
        self.path_root = path_root
        self.force = force
        self.expected_size = expected_size
        self.expected_hash = expected_hash
        self.stream: typing.Optional[io.BufferedWriter] = None
        self.hash: typing.Optional["hashlib._Hash"] = None

    def on_begin(self, manager: task.Manager) -> int:
        """Opens the local file before starting the download.

        If the file exists, this function opens it in append mode
        and returns its size in bytes.

        Args:
            manager (task.Manager): The task manager for reporting updates.

        Returns:
            int: Number of bytes already downloaded.
        """
        file_path = (
            self.path_root / self.path_id
            if self.suffix is None
            else utilities.path_with_suffix(self.path_root / self.path_id, self.suffix)
        )
        download_path = utilities.path_with_suffix(file_path, constants.DOWNLOAD_SUFFIX)
        if self.force:
            self.stream = open(download_path, "wb")
            if self.expected_hash is not None:
                self.hash = utilities.new_hash()
            return 0
        if file_path.is_file():
            size = (
                file_path.stat().st_size
                if self.expected_size is None
                else self.expected_size
            )
            manager.send_message(
                Progress(
                    path_id=self.path_id,
                    initial_bytes=size,
                    current_bytes=size,
                    final_bytes=size,
                    complete=True,
                )
            )
            return -1
        if download_path.is_file():
            if self.expected_hash is not None:
                self.hash = utilities.hash_file(
                    path=download_path, chunk_size=constants.CHUNK_SIZE
                )
            self.stream = open(download_path, "ab")
            size = download_path.stat().st_size
            manager.send_message(
                Progress(
                    path_id=self.path_id,
                    initial_bytes=size,
                    current_bytes=size,
                    final_bytes=size,
                    complete=False,
                )
            )
            return size
        self.stream = open(download_path, "wb")
        if self.expected_hash is not None:
            self.hash = utilities.new_hash()
        return 0

    def on_range_failed(self, manager: task.Manager):
        """Re-opens the file in write mode.

        Args:
            manager (task.Manager): The task manager for reporting updates.
        """
        assert self.stream is not None
        self.stream.close()
        file_path = (
            self.path_root / self.path_id
            if self.suffix is None
            else utilities.path_with_suffix(self.path_root / self.path_id, self.suffix)
        )
        download_path = utilities.path_with_suffix(file_path, constants.DOWNLOAD_SUFFIX)
        size = download_path.stat().st_size
        manager.send_message(
            Progress(
                path_id=self.path_id,
                initial_bytes=-size,
                current_bytes=-size,
                final_bytes=-size,
                complete=False,
            )
        )
        self.stream = open(download_path, "wb")
        if self.expected_hash is not None:
            self.hash = utilities.new_hash()

    def on_response_ready(
        self, response: requests.Response, manager: task.Manager
    ) -> None:
        """Iterates over the file chunks and writes them to the file.

        Args:
            response (requests.Response): HTTP response object.
            manager (task.Manager): The task manager for reporting updates.
        """
        assert self.stream is not None
        for chunk in response.iter_content(constants.CHUNK_SIZE):
            self.stream.write(chunk)
            if self.hash is not None:
                self.hash.update(chunk)
            size = len(chunk)
            manager.send_message(
                Progress(
                    path_id=self.path_id,
                    initial_bytes=0,
                    current_bytes=size,
                    final_bytes=size,
                    complete=False,
                )
            )
        response.close()
        self.on_end(manager=manager)

    def on_end(self, manager: task.Manager):
        """Checks the hash and closes the file.

        Args:
            manager (task.Manager): The task manager for reporting updates.

        Raises:
            exception.HashMismatch: if the provided and effective hashes are different.
            exception.SizeMismatch: if the provided and effective sizes are different.
        """
        if self.stream is not None:
            self.stream.close()
            if self.hash is not None:
                assert self.expected_hash is not None
                hash = self.hash.hexdigest()
                if hash != self.expected_hash:
                    raise exception.HashMismatch(self.path_id, self.expected_hash, hash)
            file_path = (
                self.path_root / self.path_id
                if self.suffix is None
                else utilities.path_with_suffix(
                    self.path_root / self.path_id, self.suffix
                )
            )
            download_path = utilities.path_with_suffix(
                file_path, constants.DOWNLOAD_SUFFIX
            )
            if self.expected_size is not None:
                size = download_path.stat().st_size
                if size != self.expected_size:
                    raise exception.SizeMismatch(self.path_id, self.expected_size, size)
            download_path.replace(file_path)
            manager.send_message(
                Progress(
                    path_id=self.path_id,
                    initial_bytes=0,
                    current_bytes=0,
                    final_bytes=0,
                    complete=True,
                )
            )



"""Binaary description of all formats using numpy dtypes."""

from __future__ import annotations

import numpy

DVS_DTYPE: numpy.dtype = numpy.dtype(
    [("t", "<u8"), ("x", "<u2"), ("y", "<u2"), ("p", "<u1")]
)
"""Data type for DVS files.

Each event uses 13 bytes, with no padding between events. The fields are packed in the following order:

- ``t``: 8 bytes unsigned integer, little endian (timestamp in µs)
- ``x``: 2 bytes unsigned integer, little endian (pixel position in the x direction in the range [0, width - 1], 0 is left)
- ``y``: 2 bytes unsigned integer, little endian (pixel position in the y direction in the range [0, height - 1], 0 is bottom)
- ``p``: 1 byte unsigned integer (polarity, 0 is OFF, 1 is ON)
"""


def aps_dtype(width: int, height: int) -> numpy.dtype:
    """Data type for APS files.

    Each frame uses 56 + (width * height * 2) bytes, with no padding between frames. Fields that are not available (for instance if the sensor does not record the frame start) are set to the maximum possible value (for instance 2 ** 64 - 1 for 8 bytes integer). The fields are packed in the following order:

    - ``t``: 8 bytes unsigned integer, little endian (timestamp in µs)
    - ``begin_t``: 8 bytes unsigned integer, little endian (AEDAT4 doc?)
    - ``end_t``: 8 bytes unsigned integer, little endian (AEDAT4 doc?)
    - ``exposure_begin_t``: 8 bytes unsigned integer, little endian (AEDAT4 doc?)
    - ``exposure_end_t``: 8 bytes unsigned integer, little endian (AEDAT4 doc?)
    - ``format``: 8 bytes unsigned integer, little endian (AEDAT4 doc?)
    - ``width``: 2 bytes unsigned integer, little endian (frame width in pixels)
    - ``height``: 2 bytes unsigned integer, little endian (frame height in pixels)
    - ``offset_x``: 2 bytes signed integer, little endian (x offset in pixels, relative to the sensor left edge)
    - ``offset_y``: 2 bytes signed integer, little endian (y offset in pixels, relative to the sensor bottom edge)
    - ``pixels``: width * height * 2 bytes, packed little endian unsigned integers (luminance values)
    """
    return numpy.dtype(
        [
            ("t", "<u8"),
            ("begin_t", "<u8"),
            ("end_t", "<u8"),
            ("exposure_begin_t", "<u8"),
            ("exposure_end_t", "<u8"),
            ("format", "<u8"),
            ("width", "<u2"),
            ("height", "<u2"),
            ("offset_x", "<i2"),
            ("offset_y", "<i2"),
            ("pixels", (numpy.dtype("<u2"), (width, height))),
        ]
    )


IMU_DTYPE: numpy.dtype = numpy.dtype(
    [
        ("t", "<u8"),
        ("temperature", "<f4"),
        ("accelerometer_x", "<f4"),
        ("accelerometer_y", "<f4"),
        ("accelerometer_z", "<f4"),
        ("gyroscope_x", "<f4"),
        ("gyroscope_y", "<f4"),
        ("gyroscope_z", "<f4"),
        ("magnetometer_x", "<f4"),
        ("magnetometer_y", "<f4"),
        ("magnetometer_z", "<f4"),
    ]
)
"""Data type for IMU files.

Each IMU sample uses 48 bytes, with no padding between samples. Fields that are not available (for instance, if the IMU has no magnetometer) are set to :py:attr:`numpy.NAN`. The fields are packed in the following order:

- ``t``: 8 bytes unsigned integer, little endian (timestamp in µs)
- ``temperature``: single-precision floating point number (IEEE 754) ()
- ``accelerometer_x``: single-precision floating point number (IEEE 754) (x-axis acceleration in m/s²)
- ``accelerometer_y``: single-precision floating point number (IEEE 754) (y-axis acceleration in m/s²)
- ``accelerometer_z``: single-precision floating point number (IEEE 754) (z-axis acceleration in m/s²)
- ``gyroscope_x``: single-precision floating point number (IEEE 754) (x-axis angular velocity in AEDAT4 doc?)
- ``gyroscope_y``: single-precision floating point number (IEEE 754) (y-axis angular velocity in AEDAT4 doc?)
- ``gyroscope_z``: single-precision floating point number (IEEE 754) (z-axis angular velocity in AEDAT4 doc?)
- ``magnetometer_x``: single-precision floating point number (IEEE 754) (x-axis angular position in AEDAT4 doc?)
- ``magnetometer_y``: single-precision floating point number (IEEE 754) (y-axis angular position in AEDAT4 doc?)
- ``magnetometer_z``: single-precision floating point number (IEEE 754) (z-axis angular position in AEDAT4 doc?)
"""



from __future__ import annotations

import argparse
import logging
import multiprocessing
import pathlib
import pkgutil
import sys
import typing

from . import (check, configuration, constants, display, formats, install_mode,
               json_index_tasks, remote, task)

dirname = pathlib.Path(__file__).resolve().parent


def check_positive(value: str):
    value_as_int = int(value)
    if value_as_int <= 0:
        raise argparse.ArgumentTypeError(f"{value} is not a positive integer")
    return value_as_int


def add_common_arguments(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--configuration",
        "-c",
        default="undr.toml",
        help="UNDR configuration file path",
    )
    parser.add_argument(
        "--timeout", "-t", default=None, type=float, help="Socket timeout in seconds"
    )
    parser.add_argument(
        "--workers",
        "-w",
        type=check_positive,
        default=multiprocessing.cpu_count() * 2,
        help="Number of parallel processes",
    )
    parser.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="Download server files even if they already exist locally",
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Download UNDR datasets listed on a configuration file",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("--log-directory", help="write log files to this directory")
    subparsers = parser.add_subparsers(dest="command")
    init_parser = subparsers.add_parser(
        "init", help="Generate a default undr.toml file"
    )
    init_parser.add_argument(
        "--configuration",
        "-c",
        default="undr.toml",
        help="UNDR configuration file path",
    )
    install_parser = subparsers.add_parser("install", help="Provision the datasets")
    add_common_arguments(install_parser)
    bibtex_parser = subparsers.add_parser(
        "bibtex", help="Generate a BibTeX file referencing all the datasets"
    )
    bibtex_parser.add_argument(
        "output",
        help='Path of the output file (for instance output.bib) or "-" to print to stdout',
    )
    add_common_arguments(bibtex_parser)
    doctor_parser = subparsers.add_parser(
        "doctor",
        help="Read all the files and check their invariants (timestamps order, spatial coordinates range...)",
    )
    add_common_arguments(doctor_parser)
    check_conformance_parser = subparsers.add_parser(
        "check-conformance",
        help="Inspect a directory before uploading it to an UNDR server",
    )
    check_conformance_parser.add_argument("path", help="Path of the local directory")
    check_conformance_parser.add_argument(
        "--skip-format-index",
        "-s",
        action="store_true",
        help="Do not format -index.json files",
    )
    args = parser.parse_args()

    log_directory: typing.Optional[pathlib.Path] = None
    if args.log_directory is not None:
        log_directory = pathlib.Path(args.log_directory)
        log_directory.mkdir(parents=True, exist_ok=True)
        logging.basicConfig(
            filename=str(log_directory / "main.log"),
            encoding="utf-8",
            level=logging.DEBUG,
            format="%(asctime)s %(message)s",
        )

    if args.command == "init":
        target = pathlib.Path(args.configuration)
        if target.is_file():
            print(display.format_error(f"{target} already exists"))
        else:
            undr_default = pkgutil.get_data("undr", "undr_default.toml")
            assert undr_default is not None
            with open(target, "wb") as target_file:
                target_file.write(undr_default)

    if args.command == "install":
        undr_configuration = configuration.configuration_from_path(
            pathlib.Path(args.configuration)
        )
        if args.timeout is not None:
            for (
                dataset_settings
            ) in undr_configuration.name_to_dataset_settings.values():
                dataset_settings.timeout = args.timeout
        try:
            undr_configuration.install(
                show_display=True,
                workers=args.workers,
                force=args.force,
                log_directory=log_directory,
            )
        except KeyboardInterrupt:
            pass

    if args.command == "bibtex":
        undr_configuration = configuration.configuration_from_path(
            pathlib.Path(args.configuration)
        )
        if args.timeout is not None:
            for (
                dataset_settings
            ) in undr_configuration.name_to_dataset_settings.values():
                dataset_settings.timeout = args.timeout
        try:
            content = undr_configuration.bibtex(
                show_display=args.output != "-",
                workers=args.workers,
                force=args.force,
                log_directory=log_directory,
                bibtex_timeout=constants.DEFAULT_TIMEOUT
                if args.timeout is None
                else args.timeout,
            )
            if args.output == "-":
                print(content)
            else:
                with open(args.output, "w") as output:
                    output.write(content)
                print(f"bibliography written to {args.output}")
        except KeyboardInterrupt:
            pass

    if args.command == "doctor":
        undr_configuration = configuration.configuration_from_path(
            pathlib.Path(args.configuration)
        )
        if args.timeout is not None:
            for (
                dataset_settings
            ) in undr_configuration.name_to_dataset_settings.values():
                dataset_settings.timeout = args.timeout
        error: typing.Optional[check.Error] = None
        for message in undr_configuration.map(
            switch=formats.Switch(
                handle_aps=check.handle_aps,
                handle_dvs=check.handle_dvs,
                handle_imu=check.handle_imu,
                handle_other=check.handle_other,
            )
        ):
            error = message
            break
        if error is None:
            print(display.format_info("No errors detected"))
        else:
            print(display.format_error(f"{error.path_id}: {error.message}"))

    if args.command == "check-conformance":
        conformance_error: typing.Optional[str] = None
        try:
            path = pathlib.Path(args.path)
            check.structure_recursive(path)
            index_status = configuration.IndexStatus(
                dataset_settings=None,  # type: ignore
                current_index_files=0,
                final_index_files=1,
                server=remote.NullServer(),
                selector=json_index_tasks.Selector(),
                downloaded_and_processed=True,
            )
            with display.Display(
                statuses=[
                    display.Status.from_path_id_and_mode(
                        path_id=pathlib.PurePosixPath(path.name),
                        dataset_mode=install_mode.Mode.RAW,
                    ),
                ],
                output_interval=constants.CONSUMER_POLL_PERIOD,
                download_speed_samples=constants.SPEED_SAMPLES,
                process_speed_samples=constants.SPEED_SAMPLES,
                download_tag=display.Tag(label="download", icon="↓"),
                process_tag=display.Tag(label="process", icon="⚛"),
            ) as terminal_display, task.ProcessManager() as manager:
                manager.schedule(
                    task=json_index_tasks.Index(
                        path_root=path.parent,
                        path_id=pathlib.PurePosixPath(path.name),
                        server=index_status.server,
                        selector=json_index_tasks.Selector(),
                        priority=0,
                        force=False,
                        directory_doi=False,
                    ),
                    priority=0,
                )
                for message in manager.messages():
                    if isinstance(message, task.WorkerException):
                        raise Exception(str(message.traceback_exception))
                    terminal_display.push(message)
                    indexing_complete, status = index_status.push(message=message)
                    if indexing_complete and status is not None:
                        manager.schedule(
                            task=check.CheckLocalDirectoryRecursive(
                                path_root=path.parent,
                                path_id=pathlib.PurePosixPath(path.name),
                                priority=1,
                            ),
                            priority=1,
                        )
            if not args.skip_format_index:
                check.format_index_recursive(
                    path=path,
                    handle_path=lambda path: print(
                        display.format_info(f"Formatted {path}")
                    ),
                )
        except KeyboardInterrupt:
            conformance_error = "Interrupted"
        except:
            conformance_error = str(sys.exc_info()[1])
        if conformance_error is None:
            print(display.format_info("No errors detected"))
        else:
            print(display.format_error(conformance_error))



"""BibTeX downloading and parsing."""

from __future__ import annotations

import logging

import requests


def from_doi(doi: str, pretty: bool, timeout: float) -> str:
    """Downloads and formats BibTeX from https://dx.doi.org/.

    Args:
        doi (str): Digital object identifier (DOI) string starting with ``10.``.
        pretty (bool): Whether to correct line breaks and indents.
            Otherwise, this function returns the raw reponse from https://dx.doi.org/.
        timeout (float): Request timeout in seconds.

    Raises:
        :py:class:`requests.exceptions.HTTPError`: if a network error occurs (unreachable server, timeout...).

    Returns:
        str: BibTeX entry for the given DOI.
    """
    logging.debug(f"request application/x-bibtex from https://dx.doi.org/{doi}")
    response = requests.get(
        f"https://dx.doi.org/{doi}",
        timeout=timeout,
        headers={"Accept": "application/x-bibtex; charset=utf-8"},
    )
    response.raise_for_status()
    if pretty:
        bibtex = ""
        new_line = True
        depth = 0
        for character in response.text:
            if new_line:
                if not character.isspace():
                    new_line = False
                    bibtex += " " * ((depth - 1 if character == "}" else depth) * 4)
            if character == "{":
                depth += 1
                bibtex += "{"
            elif character == "}":
                depth -= 1
                bibtex += "}"
            elif character == "\n":
                new_line = True
                bibtex += "\n"
            elif character.isspace():
                if not new_line:
                    bibtex += character
            else:
                bibtex += character
        if not bibtex.endswith("\n"):
            bibtex += "\n"
        return bibtex
    return response.text



"""Custom runtime exceptions.

Direct calls to :py:class:`Exception` in other parts of the codebase indicate a logic error.
"""

import pathlib


class HashMismatch(Exception):
    """Raised if the hash read from the index is different from the hash calculated from the file bytes.

    Args:
        path_id (pathlib.PurePosixPath): The path ID of the resource with a hash mismatch.
        expected_hash (str): The hash read from the index.
        hash (str): The hash calculated from the bytes effectively processed.
    """

    def __init__(self, path_id: pathlib.PurePosixPath, expected_hash: str, hash: str):
        super().__init__(
            f'hash mismatch for "{path_id}" (expected "{expected_hash}", got "{hash}")'
        )


class SizeMismatch(Exception):
    """Raised if the size read from the index is different from number of bytes in the decompressed file.

    Args:
        path_id (pathlib.PurePosixPath): The path ID of the resource with a size mismatch.
        expected_size (int): The size read from the index.
        size (int): The effective decompressed size in bytes.
    """

    def __init__(self, path_id: pathlib.PurePosixPath, expected_size: int, size: int):
        super().__init__(
            f'size mistmatch for "{path_id}" (expected "{expected_size}", got "{size}")'
        )



"""Terminal progress bar."""

from __future__ import annotations

import collections
import copy
import dataclasses
import itertools
import logging
import os
import pathlib
import sys
import threading
import time
import types
import typing

from . import decode, install_mode, json_index_tasks, remote, utilities

ANSI_COLORS_ENABLED = os.getenv("ANSI_COLORS_DISABLED") is None


def format_bold(message: str) -> str:
    """Surrounds the message with ANSI escape characters for bold formatting.

    Args:
        message (str): A message to be displayed in a terminal.

    Returns:
        str: The message surrounded with ANSI escape characters, or the original message if the environment variable ``ANSI_COLORS_DISABLED`` is set.
    """
    if ANSI_COLORS_ENABLED:
        return f"\033[1m{message}\033[0m"
    return message


def format_dim(message: str) -> str:
    """Surrounds the message with ANSI escape characters for dim formatting.

    Args:
        message (str): A message to be displayed in a terminal.

    Returns:
        str: The message surrounded with ANSI escape characters, or the original message if the environment variable ``ANSI_COLORS_DISABLED`` is set.
    """
    if ANSI_COLORS_ENABLED:
        return f"\033[2m{message}\033[0m"
    return message


def format_info(message: str) -> str:
    """Adds a kangaroo icon in front of the message.

    Args:
        message (str): A message to be displayed in a terminal.

    Returns:
        str: The message with a prefix.
    """
    return f"🦘 {format_bold(message)}"


def format_error(message: str) -> str:
    """Adds a red cross mark in front of the message.

    Args:
        message (str): A message to be displayed in a terminal.

    Returns:
        str: The message with a prefix.
    """
    return f"❌ {message}"


def progress_bar(width: int, progress: typing.Optional[tuple[float, float]]) -> str:
    """Generates a progress bar compatible with terminals.

    Args:
        width (int): The progress bar width in characters.
        progress (typing.Optional[tuple[float, float]]): None yields an indeterminate progress bar, a tuple returns a two-levels progress bar (bottom and top). The tuple values must be in the range ``[0, 1]``. Use the same value twice to generate a simple (one-level) progress bar.

    Returns:
        str: The progress bar as a string, without line breaks.
    """

    width = max(width, 3)
    if progress is None:
        return "|{}|".format("░" * (width - 2))
    top = max(0.0, min(1.0, progress[0]))
    bottom = max(0.0, min(1.0, progress[1]))
    top_fill = round((width - 2) * top)
    bottom_fill = round((width - 2) * bottom)
    if top_fill < bottom_fill:
        return "|{}{}{}|".format(
            "█" * top_fill,
            "▄" * (bottom_fill - top_fill),
            "–" * (width - 2 - bottom_fill),
        )
    if top_fill > bottom_fill:
        return "|{}{}{}|".format(
            "█" * bottom_fill,
            "▀" * (top_fill - bottom_fill),
            "–" * (width - 2 - top_fill),
        )
    return "|{}{}|".format("█" * top_fill, "–" * (width - 2 - top_fill))


@dataclasses.dataclass
class DisplayProgress:
    """Represents download or process progress."""

    initial_bytes: int
    """Number of bytes already downloaded or processed when the action started.
    """

    current_bytes: int
    """Number of bytes currently downloaded or processed.
    """

    final_bytes: int
    """Total number of bytes to download or process.
    """


@dataclasses.dataclass
class Status:
    """Keeps track of download and process progress for a dataset."""

    path_id: pathlib.PurePosixPath
    """Path ID of the dataset's base directory.
    """

    mode: install_mode.Mode
    """Dataset installation mode.
    """

    indexing: bool
    """Whether the dataset has been indexed.
    """

    current_index_files: int
    """Number of index files downloaded so far.
    """

    final_index_files: int
    """Total number of index files.

    This number may increase as more index files are discovered while indexing.
    """

    download: DisplayProgress
    """Represents download progress.

    Ignored if the mode is :py:attr:`undr.install_mode.Mode.REMOTE`.
    """

    process: DisplayProgress
    """Represents process progress

    Ignored if the mode is :py:attr:`undr.install_mode.Mode.REMOTE` or :py:attr:`undr.install_mode.Mode.LOCAL`.
    """

    @classmethod
    def from_path_id_and_mode(
        cls, path_id: pathlib.PurePosixPath, dataset_mode: install_mode.Mode
    ):
        """Initializes a status from a path ID and a mode.

        Args:
            path_id (pathlib.PurePosixPath): The dataset's base directory.
            dataset_mode (install_mode.Mode): The installation mode.

        Returns:
            Status: Default initial status.
        """
        return cls(
            path_id=path_id,
            mode=dataset_mode,
            indexing=True,
            current_index_files=0,
            final_index_files=1,
            download=DisplayProgress(
                initial_bytes=0,
                current_bytes=0,
                final_bytes=0,
            ),
            process=DisplayProgress(
                initial_bytes=0,
                current_bytes=0,
                final_bytes=0,
            ),
        )

    def speeds(self, previous_status: "Status", interval: float) -> tuple[float, float]:
        """Calculates download and process speeds.

        Args:
            previous_status (Status): Past status used to calculate progress variation.
            interval (float): Time since the previous status in seconds.

        Returns:
            tuple[float, float]: download speed and process speed in bytes per second.
        """
        return (
            (self.download.current_bytes - self.download.initial_bytes)
            - (
                previous_status.download.current_bytes
                - previous_status.download.initial_bytes
            )
        ) / interval, (
            (self.process.current_bytes - self.process.initial_bytes)
            - (
                previous_status.process.current_bytes
                - previous_status.process.initial_bytes
            )
        ) / interval

    def label(self) -> str:
        """Returns the dataset's name."""
        return "/".join(self.path_id.parts)

    def progress_and_representation(
        self, download_tag: "Tag", process_tag: "Tag"
    ) -> tuple[typing.Optional[tuple[float, float]], str]:
        """Returns download and process progress in bytes and a string representation of these values.

        Args:
            download_tag (Tag): Icon and label for the download action.
            process_tag (Tag): Icon and label for the process action.

        Returns:
            tuple[tuple[float, float], str]: The first entry is None if the dataset is being indexed. Otherwise, the first entry is the progress in bytes. The second entry is the string representation of progress, which is always not None.
        """
        if self.indexing:
            return None, f"index {self.current_index_files} / {self.final_index_files}"
        representation = f"{download_tag.icon} "
        if self.download.current_bytes == self.download.final_bytes:
            representation += utilities.size_to_string(self.download.current_bytes)
        else:
            representation += f"{utilities.size_to_string(self.download.current_bytes)} / {utilities.size_to_string(self.download.final_bytes)}"
        download_progress = (
            1.0
            if self.download.final_bytes == 0
            else self.download.current_bytes / self.download.final_bytes
        )
        if self.mode == install_mode.Mode.RAW:
            representation += f" | {process_tag.icon} "
            if self.process.current_bytes == self.process.final_bytes:
                representation += utilities.size_to_string(self.process.current_bytes)
            else:
                representation += f"{utilities.size_to_string(self.process.current_bytes)} / {utilities.size_to_string(self.process.final_bytes)}"
            progress = (
                download_progress,
                1.0
                if self.process.final_bytes == 0
                else self.process.current_bytes / self.process.final_bytes,
            )
        else:
            progress = (download_progress, download_progress)
        return progress, representation

    def complete(self) -> bool:
        """Checks whether actions are complete for this dataset.

        Returns:
            bool: Whether all actions are complete.
        """
        return (
            not self.indexing
            and self.download.current_bytes == self.download.final_bytes
            and self.process.current_bytes == self.process.final_bytes
        )

    def is_parent_of(self, path_id: pathlib.PurePosixPath) -> bool:
        """Checks whether the dataset represented by this status is a parent of the given resource.

        This can be used to assign messages from a given resource to the right dataset.

        Args:
            path_id (pathlib.PurePosixPath): The path ID of the resource that may be a child of this dataset.

        Returns:
            bool: Whether this dataset is a parent of the resource.
        """
        if len(self.path_id.parts) > len(path_id.parts):
            return False
        for part, other_part in zip(self.path_id.parts, path_id.parts):
            if part != other_part:
                return False
        return True


@dataclasses.dataclass
class Tag:
    """Label and icon representing an action in a terminal."""

    label: str
    """The action label.
    """

    icon: str
    """The action's icon, typically a single non-ASCII UTF-8 character.
    """


class Speedometer:
    """Measures speed with multiple samples and a sliding window.

    Args:
        maximum_samples (int): Number of samples. Fewer samples are used for the first few speed estimations, until that number is reached.
    """

    def __init__(self, maximum_samples: int):
        assert maximum_samples > 0
        self.samples: collections.deque[float] = collections.deque()
        self.maximum_samples = maximum_samples

    def __repr__(self) -> str:
        return f"{self.__class__}({self.__dict__})"

    def add_sample(self, sample: float):
        """Passes a new sample to the speedometer.

        Args:
            sample (float): Speed sample in bytes per second.
        """
        if len(self.samples) >= self.maximum_samples:
            self.samples.popleft()
        self.samples.append(sample)

    def value(self) -> float:
        """Current speed value in bytes per second.

        Returns:
            float: Mean value of the samples.
        """

        return sum(self.samples) / len(self.samples)


def speeds(
    previous_statuses: list[Status],
    statuses: list[Status],
    interval: float,
) -> tuple[float, float]:
    """Calculates the total speed for multiple datasets.

    Args:
        previous_statuses (list[Status]): Past statuses used to calculate progress variation.
        statuses (list[Status]): Current statuses, must be the same length as `previous_statuses`.
        interval (float): Time since the previous statuses in seconds.

    Returns:
        tuple[float, float]: Overall download speed and process speed in bytes per second.
    """
    total_download = 0
    total_process = 0
    for status, previous_status in zip(statuses, previous_statuses):
        download, process = status.speeds(previous_status, interval)
        total_download += download
        total_process += process
    return total_download, total_process


class Display:
    """Display collects update messages and shows progress in a terminal.

    Args:
        statuses (list[Status]): Initial statuses for all datasets.
        output_interval (float): Time between refreshes in seconds.
        download_speed_samples (int): Size of the sliding window for download speed calculations.
        process_speed_samples (int): Size of the sliding window for process speed calculations.
        download_tag (Tag): Label and icon for the download action.
        process_tag (Tag): Label and icon for the process action.
    """

    def __init__(
        self,
        statuses: list[Status],
        output_interval: float,
        download_speed_samples: int,
        process_speed_samples: int,
        download_tag: Tag,
        process_tag: Tag,
    ):
        self.output_interval = output_interval
        self.download_speedometer = Speedometer(maximum_samples=download_speed_samples)
        self.process_speedometer = Speedometer(maximum_samples=process_speed_samples)
        self.download_tag = download_tag
        self.process_tag = process_tag
        self.names = tuple(str(status.path_id) for status in statuses)
        self.name_width = max(len(name) for name in self.names) + 4
        self.show_process_speed = any(
            status.mode == install_mode.Mode.RAW for status in statuses
        )
        self.running = True
        self.finalize = True
        self.previous_statuses = copy.deepcopy(statuses)
        self.begin = time.monotonic()
        self.previous_lines: list[str] = []
        self.message_queue: collections.deque[
            typing.Union[
                decode.Progress,
                remote.Progress,
                json_index_tasks.IndexLoaded,
                json_index_tasks.DirectoryScanned,
            ]
        ] = collections.deque()
        self.worker = threading.Thread(target=self.target)
        self.worker.daemon = True
        self.worker.start()
        logging.debug(f"{self.__dict__}")

    def push(self, message: typing.Any):
        """Processes update messages.

        Ignores messages that are not :py:class:`undr.decode.Progress`, :py:class:`undr.remote.Progress`, :py:class:`undr.json_index_tasks.IndexLoaded` or :py:class:`undr.json_index_tasks.DirectoryScanned`.

        Args:
            message (typing.Any): Message from a worker.
        """
        if isinstance(
            message,
            (
                decode.Progress,
                remote.Progress,
                json_index_tasks.IndexLoaded,
                json_index_tasks.DirectoryScanned,
            ),
        ):
            self.message_queue.append(message)

    def messages(self, statuses: list[Status]) -> list[Status]:
        """Consumes messages until the queue is empty.

        This function consumes messages until the queue is empty, not closed. More messages are likely to be queued after this function returns.

        Args:
            statuses (list[Status]): Current statuses, will be modified in-place. Use :py:func:`copy.deepcopy` to preserve the original list.

        Raises:
            RuntimeError: if a message in the queue is not :py:class:`undr.decode.Progress`, :py:class:`undr.remote.Progress`, :py:class:`undr.json_index_tasks.IndexLoaded` or :py:class:`undr.json_index_tasks.DirectoryScanned`.

        Returns:
            list[Status]: Updated statuses.
        """
        while True:
            try:
                message = self.message_queue.popleft()
                status: typing.Optional[Status] = None
                for candidate_status in statuses:
                    if candidate_status.is_parent_of(message.path_id):
                        status = candidate_status
                        break
                assert status is not None
                if isinstance(message, decode.Progress):
                    if message.initial_bytes < 0:
                        status.process.initial_bytes += message.initial_bytes
                        status.process.current_bytes += message.current_bytes
                    elif message.initial_bytes == 0:
                        status.process.current_bytes += message.current_bytes
                elif isinstance(message, remote.Progress):
                    if message.initial_bytes < 0 and not status.indexing:
                        status.download.initial_bytes += message.initial_bytes
                        status.download.current_bytes += message.current_bytes
                    if message.initial_bytes == 0:
                        status.download.current_bytes += message.current_bytes
                elif isinstance(message, json_index_tasks.IndexLoaded):
                    status.final_index_files += message.children
                elif isinstance(message, json_index_tasks.DirectoryScanned):
                    status.current_index_files += 1
                    status.download.initial_bytes += (
                        message.index_bytes.initial + message.download_bytes.initial
                    )
                    status.download.current_bytes += (
                        message.index_bytes.initial + message.download_bytes.initial
                    )
                    status.download.final_bytes += (
                        message.index_bytes.final + message.download_bytes.final
                    )
                    status.process.initial_bytes += message.process_bytes.initial
                    status.process.current_bytes += message.process_bytes.initial
                    status.process.final_bytes += message.process_bytes.final
                    if status.current_index_files == status.final_index_files:
                        status.indexing = False
                else:
                    raise RuntimeError(
                        f"unexpected message {message.__class__} ({message})"
                    )
            except IndexError:
                break
        return statuses

    def target(self):
        """Worker thread implementation."""
        next_dispatch = time.monotonic()
        self.output(
            statuses=self.previous_statuses,
            average=False,
            download_speed=0.0,
            process_speed=0.0,
        )
        previous_speed_estimation = self.begin
        while self.running:
            statuses = self.messages(copy.deepcopy(self.previous_statuses))
            now = time.monotonic()
            download_speed, process_speed = speeds(
                self.previous_statuses, statuses, now - previous_speed_estimation
            )
            previous_speed_estimation = now
            self.download_speedometer.add_sample(download_speed)
            self.process_speedometer.add_sample(process_speed)
            self.output(
                statuses=statuses,
                average=False,
                download_speed=self.download_speedometer.value(),
                process_speed=self.process_speedometer.value(),
            )
            self.previous_statuses = statuses
            next_dispatch += self.output_interval
            now = time.monotonic()
            if next_dispatch > now:
                time.sleep(next_dispatch - now)
        if self.finalize:
            statuses = self.messages(copy.deepcopy(self.previous_statuses))
            download_speed = 0.0
            process_speed = 0.0
            duration = time.monotonic() - self.begin
            for status in statuses:
                download_speed += (
                    status.download.current_bytes - status.download.initial_bytes
                ) / duration
                process_speed += (
                    status.process.current_bytes - status.process.initial_bytes
                ) / duration
            self.output(
                statuses=statuses,
                average=True,
                download_speed=download_speed,
                process_speed=process_speed,
            )
            self.previous_statuses = statuses

    def close(self):
        """Terminates the worker thread.

        This function is called automatically if display is used as a context manager.
        """
        self.running = False
        self.worker.join()

    def __enter__(self) -> "Display":
        """Enables the use of the "with" statement.

        Returns:
            Display: A display context that calls :py:meth:`close` on exit.
        """
        return self

    def __exit__(
        self,
        type: typing.Optional[typing.Type[BaseException]],
        value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ):
        """Enables the use of the "with" statement.

        Args:
            type (typing.Optional[typing.Type[BaseException]]): None if the context exits without an exception, and the raised exception's class otherwise.
            value (typing.Optional[BaseException]): None if the context exits without an exception, and the raised exception otherwise.
            traceback (typing.Optional[types.TracebackType]): None if the context exits without an exception, and the raised exception's traceback otherwise.
        """
        logging.debug(f"display exit with error type {type}")
        if type is not None:
            self.finalize = False
        self.close()

    def time_left(
        self, statuses: list[Status], download_speed: float, process_speed: float
    ) -> typing.Optional[float]:
        """Estimates the time left to complete the download and process actions.

        Args:
            statuses (list[Status]): Current datasets statuses.
            download_speed (float): Current speed in bytes per second.
            process_speed (float): Current speed in bytes per second.

        Returns:
            typing.Optional[float]: Estimated time left in second or None if at least one dataset is still being indexed or if the download and process speeds are zero.
        """
        download_left = 0
        process_left = 0
        for status in statuses:
            if status.indexing:
                return None
            download_left += status.download.final_bytes - status.download.current_bytes
            if status.mode == install_mode.Mode.RAW:
                process_left += (
                    status.process.final_bytes - status.process.current_bytes
                )
        if download_speed > 0 and process_speed > 0:
            return max(download_left / download_speed, process_left / process_speed)
        if download_speed > 0:
            return download_left / download_speed
        if process_speed > 0:
            return process_left / process_speed
        return None

    def output(
        self,
        statuses: list[Status],
        average: bool,
        download_speed: float,
        process_speed: float,
    ):
        """Called by the worker to generate the terminal text.

        Args:
            statuses (list[Status]): Current datasets statuses.
            average (bool): Whether the provided speeds are averages, typically used after all the actions are complete.
            download_speed (float): Current or average download speed in bytes per second.
            process_speed (float): Current or average process speed in bytes per second.
        """
        progresses_and_representations = tuple(
            status.progress_and_representation(self.download_tag, self.process_tag)
            for status in statuses
        )
        representation_width = max(
            len(representation) for _, representation in progresses_and_representations
        )
        progress_bar_width = (
            os.get_terminal_size().columns
            - self.name_width
            - (representation_width + 2)
        )
        labels = tuple(
            f"{'✓' if status.complete() else '⋅'} {status.path_id}"
            for status in statuses
        )
        time_left: typing.Optional[float] = None
        overview = ""
        if average:
            overview += "average: "
        else:
            time_left = self.time_left(statuses, download_speed, process_speed)
        overview += f"{self.download_tag.icon} {self.download_tag.label} {utilities.speed_to_string(round(download_speed))}"
        if self.show_process_speed:
            overview += f" | {self.process_tag.icon} {self.process_tag.label} {utilities.speed_to_string(round(process_speed))}"
        if time_left is not None:
            overview += f" | ⧗ {utilities.duration_to_string(time_left)} left"
        lines = [
            *(
                f"{label:<{self.name_width}}{progress_bar(progress_bar_width, progress)} {representation} "
                for label, (progress, representation) in zip(
                    labels, progresses_and_representations
                )
            ),
            "",
            overview,
        ]
        if len(self.previous_lines) > 0:
            sys.stdout.write(f"\r\033[{len(self.previous_lines)}A")
        for line, previous_line in itertools.zip_longest(lines, self.previous_lines):
            if line == previous_line:
                sys.stdout.write(f"\033[1B")
            else:
                if line is None:
                    sys.stdout.write(f"{' ' * len(previous_line)}\n")
                elif previous_line is None:
                    sys.stdout.write(f"{line}\n")
                else:
                    sys.stdout.write(f"{' ' * len(previous_line)}\r{line}\n")
        sys.stdout.flush()
        self.previous_lines = lines



"""File formats supported by UNDR.

File types that are not listed here can still be included, downloaded, decompressed, and checked with UNDR. However, they are listed as "other files" and UNDR does not attempt to parse or load them.

To add a new file format, create and implement a derived class of :py:class:`undr.path.File` (see for instance :py:class:`ApsFile`) and add it to :py:func:`file_from_dict` and :py:class:`Switch`.
"""

from __future__ import annotations

import dataclasses
import functools
import typing

import numpy

from . import path, raw

if typing.TYPE_CHECKING:
    from . import path_directory

try:
    functools.cached_property
except:
    functools.cached_property = property


@dataclasses.dataclass(frozen=True)
class ApsFile(path.File):
    """A file that contains luminance (grey levels) frames.

    Active-pixel sensors (APS) describe, strictly speaking, any sensor with pixels that use MOSFET amplifiers.
    However, the term is commonly used to refer to the integrating pixels found in non-CCD conventional cameras.
    In the context of Neuromorphic Engineering, APS is used to describe DAVIS frames.
    """

    width: int = dataclasses.field(compare=False, hash=False)
    """Frame width (x direction) in pixels.
    """

    height: int = dataclasses.field(compare=False, hash=False)
    """Frame height (y direction) in pixels.
    """

    @functools.cached_property
    def word_size(self) -> int:
        return raw.aps_dtype(self.width, self.height).itemsize  # typr: ignore

    def packets(self) -> typing.Iterable[numpy.ndarray]:
        """Iterates over the file data.

        This function streams the file from the remote server if it is not available locally, and decompresses the file in memory if it is locally available but compressed.

        Returns:
            typing.Iterable[numpy.ndarray]: Iterator over the file's data converted into numpy arrays with dtype :py:func:`undr.raw.aps_dtype`.
        """
        dtype = numpy.dtype(raw.aps_dtype(self.width, self.height))
        for chunk in self._chunks(word_size=self.word_size):
            yield numpy.frombuffer(chunk, dtype=dtype)


@dataclasses.dataclass(frozen=True)
class DvsFile(path.File):
    """A file that contains DVS (polarity) events.

    Dynamic Vision Sensor events, often called polarity events,
    contain a timestamp, pixel coordinates, and a polarity (ON or OFF)
    that indicates whether luminance increased or decreased.
    """

    width: int = dataclasses.field(compare=False, hash=False)
    """Frame width (x direction) in pixels.
    """

    height: int = dataclasses.field(compare=False, hash=False)
    """Frame height (y direction) in pixels.
    """

    @functools.cached_property
    def word_size(self):
        return raw.DVS_DTYPE.itemsize

    def packets(self) -> typing.Iterable[numpy.ndarray]:
        """Iterates over the file data.

        This function streams the file from the remote server if it is not available locally, and decompresses the file in memory if it is locally available but compressed.

        Returns:
            typing.Iterable[numpy.ndarray]: Iterator over the file's data converted into numpy arrays with dtype :py:attr:`undr.raw.DVS_DTYPE`.
        """
        for chunk in self._chunks(word_size=self.word_size):
            yield numpy.frombuffer(chunk, dtype=raw.DVS_DTYPE)


@dataclasses.dataclass(frozen=True)
class ImuFile(path.File):
    """A file that contains IMU events.

    Inertial Measurement Unit (IMU) events are produced by an accelerometer / gyroscope / magnetometer.
    """

    @functools.cached_property
    def word_size(self):
        return raw.IMU_DTYPE.itemsize

    def packets(self) -> typing.Iterable[numpy.ndarray]:
        """Iterates over the file data.

        This function streams the file from the remote server if it is not available locally, and decompresses the file in memory if it is locally available but compressed.

        Returns:
            typing.Iterable[numpy.ndarray]: Iterator over the file's data converted into numpy arrays with dtype :py:attr:`undr.raw.IMU_DTYPE`.
        """
        for chunk in self._chunks(word_size=self.word_size):
            yield numpy.frombuffer(chunk, dtype=raw.IMU_DTYPE)


def file_from_dict(
    data: dict[str, typing.Any],
    parent: path_directory.Directory,
) -> path.File:
    """Creates a file object from a dictionary, typically loaded from an index file.

    Args:
        data (dict[str, typing.Any]): Dictionary describing the file and its properties.
        parent (path_directory.Directory): Parent directory of the file, used to generate the file system path.

    Raises:
        RuntimeError: if the file type is not supported by this function.

    Returns:
        path.File: A specialized file object.
    """
    file_attributes = path.File.attributes_from_dict(data=data, parent=parent)
    if data["properties"]["type"] == "aps":
        return ApsFile(
            **file_attributes,
            width=data["properties"]["width"],
            height=data["properties"]["height"],
        )
    if data["properties"]["type"] == "dvs":
        return DvsFile(
            **file_attributes,
            width=data["properties"]["width"],
            height=data["properties"]["height"],
        )
    if data["properties"]["type"] == "imu":
        return ImuFile(**file_attributes)
    raise RuntimeError(f"unsupported file type {type}")


SendMessage = typing.Callable[[typing.Any], None]
"""Callback channel for messages generated by a file handler during data iteration.
"""


@dataclasses.dataclass
class Switch:
    """Calls specialized file handlers while iterating a dataset.

    If a handler is None, the corresponding files are ignored by the iterator.

    Args:
        handle_aps (typing.Optional[typing.Callable[[ApsFile, SendMessage], None]]): Handler for APS files.
        handle_dvs (typing.Optional[typing.Callable[[DvsFile, SendMessage], None]]): Handler for DVS files.
        handle_imu (typing.Optional[typing.Callable[[ImuFile, SendMessage], None]]): Handler for IMU files.
        handle_other (typing.Optional[typing.Callable[[path.File, SendMessage], None]]): Handler for other files.
    """

    handle_aps: typing.Optional[typing.Callable[[ApsFile, SendMessage], None]] = None
    handle_dvs: typing.Optional[typing.Callable[[DvsFile, SendMessage], None]] = None
    handle_imu: typing.Optional[typing.Callable[[ImuFile, SendMessage], None]] = None
    handle_other: typing.Optional[
        typing.Callable[[path.File, SendMessage], None]
    ] = None

    def enabled_types(self) -> set[typing.Any]:
        """Lists the file types that have a non-None handler.

        Returns:
            set[typing.Any]: The set of file classes that will be handled.
        """
        result = set()
        if self.handle_aps is not None:
            result.add(ApsFile)
        if self.handle_dvs is not None:
            result.add(DvsFile)
        if self.handle_imu is not None:
            result.add(ImuFile)
        return result

    def handle_file(self, file: path.File, send_message: SendMessage):
        """Calls the specialized file handler for the file, if the handler is non-None.

        Args:
            file (path.File): The file to process.
            send_message (SendMessage): Callback channel for messages.

        Raises:
            RuntimeError: if the file type is not supported by this function.
        """
        if self.handle_aps is not None and isinstance(file, ApsFile):
            self.handle_aps(file, send_message)
        elif self.handle_dvs is not None and isinstance(file, DvsFile):
            self.handle_dvs(file, send_message)
        elif self.handle_imu is not None and isinstance(file, ImuFile):
            self.handle_imu(file, send_message)
        elif self.handle_other is not None and file.__class__ == path.File:
            self.handle_other(file, send_message)
        else:
            raise RuntimeError(f"unsupported file format {file.__class__}")



"""Multi-process task management."""

from __future__ import annotations

import collections
import enum
import logging
import multiprocessing
import pathlib
import pickle
import socket
import socketserver
import struct
import threading
import time
import traceback
import types
import typing

import requests

from . import constants


class Task:
    """A processing task to be performed by a worker."""

    def __repr__(self) -> str:
        return f"{self.__class__}({self.__dict__})"

    def run(self, session: requests.Session, manager: "Manager"):
        raise NotImplementedError()


class Chain(Task):
    """A sequence of tasks that must run sequentially.

    Args:
        Task (typing.Sequence[Task]): The list of tasks to run sequentially in the given order.
    """

    def __init__(self, tasks: typing.Sequence[Task]):
        self.tasks = tasks

    def run(self, session: requests.Session, manager: "Manager"):
        for task in self.tasks:
            task.run(session=session, manager=manager)


class Manager:
    """Schedules and keeps track of tasks.

    This is an abtract class, use of one of its implementations such as :py:class:`ProcessManager` to create objects.
    """

    def schedule(self, task: Task, priority: int = 1) -> None:
        """Runs a task with the given priority.

        Tasks with lower priorities are scheduled first. The maximum priority level depends on the implementation. At least two levels, 0 (highest priority) and 1, must be supported by all implementations.

        Args:
            task (Task): The task that this manager must run (possibly on a different thread).
            priority (int, optional): Priority level. Defaults to 1.
        """
        raise NotImplementedError()

    def send_message(self, message: typing.Any) -> None:
        """Queues a message in the manager's "inbox".

        A manager is responsible for collecting messages from all tasks, which are potentially running on different threads or processes, and serving thse messages in a single-threaded fashion to a reader.

        This function is meant to be called by tasks, which have access to the manager in their :py:meth:`Task.run` function.

        Args:
            message (typing.Any): Any object. Currently implemeted managers require the message to be compatible with the :py:mod:`pickle` module.
        """
        raise NotImplementedError()


class NullManager(Manager):
    """Manager placeholder that ignores messages and raises errors if one attemps to use it to schedule more tasks.

    This manager can be used with one-off tasks whose progress need not be monitored and which do not generate more tasks.
    """

    def send_message(self, message: typing.Any):
        pass


class WorkerException(Exception):
    """An exception wrapper than can be sent across threads.

    This exception captures the stack trace of the thread that raised it to improve error reporting.

    Args:
        traceback_exception (traceback.TracebackException): Traceback of the orignal exception, can be obtained with :py:meth:`traceback.TracebackException.from_exception`.
    """

    def __init__(self, traceback_exception: traceback.TracebackException):
        self.traceback_exception = traceback_exception

    def __str__(self):
        return f"\n{''.join(self.traceback_exception.format()).strip()}"


class CloseRequest:
    """Special task used to request a worker thread shutdown."""

    pass


def send_bytes(client: socket.socket, type: bytes, message: bytes):
    """Packs and sends the bytes of a type and message.

    This message encoding scheme is used internally by :py:class:`ProcessManager`.

    Args:
        client (socket.socket): TCP client used to send messages between workers and the manager.
        type (bytes): Encoded type bytes. See :py:func:`send_bytes` for a description of type encoding.
        message (bytes): Pickled message bytes.
    """
    client.sendall(struct.pack("<cQ", type, len(message)) + message)


def send_message(client: socket.socket, type: bytes, message: typing.Any):
    """Packs and sends the bytes of a type and an unencoded message.

    This message encoding scheme is used internally by :py:class:`ProcessManager`.

    Args:
        client (socket.socket): TCP client used to send messages between workers and the manager.
        type (bytes): Encoded type bytes. See :py:func:`send_bytes` for a description of type encoding.
        message (typing.Any): Any object compatible with the :py:mod:`pickle` module.
    """
    send_bytes(client=client, type=type, message=pickle.dumps(message))


def send_type(client: socket.socket, type: bytes):
    """Packs and sends the bytes of a type that does not require a message.

    Types encode the following information.

    Messages sent by a worker to the manager.

    - ``b"n"``: Reports that the worker started a task, must not have an attached message.
    - ``b"t"``: Reports that the worker completed a task and is idle, must not have an attached message.
    - ``b"m"`` Generic message that must be forwarded to the "inbox", must have an attached message
    - ``>= 0x80``: Tells the manager to spawn a new task (a worker may do this multiple times per task). The new task priority is ``type - 0x80``. This scheme supports up to 128 priority levels. The current implementation uses 2 by default.

    Messages sent by the manager to a worker.

    - ``b"t"``: Tells the worker to start a new task, must have an attached message. The task may be an instance of :py:class:`CloseRequest`, which tells the worker to shutdown.
    - ``b"m"``: Acknowledges a generic message (message to worker ``b"m"`` request).
    - ``b"s"``: Acknowledges a task message (message to worker ``>= 0x80``  request).

    Args:
        client (socket.socket): TCP client used to send messages between workers and the manager.
        type (bytes): Encoded type bytes.
    """
    client.sendall(struct.pack("<cQ", type, 0))


def receive_message(
    client: socket.socket, unpickle: bool = True
) -> tuple[bytes, typing.Any]:
    """Reads TCP bytes until enough are received to generate a full a type and message.

    Args:
        client (socket.socket): TCP client used to send messages between workers and the manager.
        unpickle (bool, optional): Whether to pass the message bytes to :py:func:`pickle.loads`. Defaults to True.

    Returns:
        tuple[bytes, typing.Any]: The type's bytes and the decoded message. The type's bytes and the raw message's bytes are returned instead if unpickle is False.
    """
    header: bytes = b""
    while True:
        length = len(header)
        header += client.recv(9 - length)
        new_length = len(header)
        if new_length == length:
            raise ConnectionResetError()
        if new_length == 9:
            break
    type, size = struct.unpack("<cQ", header)
    if size == 0:
        return type, b""
    message = bytearray(size)
    offset = 0
    while offset < size:
        read_bytes = client.recv_into(memoryview(message)[offset:])
        if read_bytes == 0:
            raise ConnectionResetError()
        offset += read_bytes
    return type, pickle.loads(message) if unpickle else message


def receive_bytes(client: socket.socket) -> tuple[bytes, bytes]:
    """Reads TCP bytes until enough are received to generate a full a type and a raw message.

    Args:
        client (socket.socket): TCP client used to send messages between workers and the manager.

    Returns:
        tuple[bytes, bytes]: The type's bytes and the raw message's bytes.
    """
    return receive_message(client=client, unpickle=False)


def receive_type(client: socket.socket, expected_type: bytes):
    """Reads TCP bytes bytes until an acknowledge type is received.

    Args:
        client (socket.socket): TCP client used to send messages between workers and the manager.
        expected_type (bytes): The execpted acknowledge type.
    """
    type, message = receive_message(client=client, unpickle=False)
    assert type == expected_type and message == b""


class ProcessManager(Manager):
    """Implements a manager that controls a pool of worker processes.

    This class is similar to :py:class:`multiprocessing.pool.Pool` but it has better support for user-initiated shutdowns (sigint / CTRL-C) and worker-initiated shutdowns (exceptions). It also supports priorities levels.

    Whenever a worker is idle, the manager scans its (the manager's) task queues in order of priority until it finds a non-empty queue, and sends the first task from that queue to the worker. Hence, tasks with lower priorities are scheduled first. However, since a task may asynchronously spawn more tasks with arbitrary priority levels, there is no guarantee that all tasks with priority 0 spawned by a program overall are executed before all tasks with priority 1. In particular, tasks are never cancelled, even if a task with a lower priority level (i.e. more urgent) becomes available while a worker is already running a task with a higher priority level (i.e. less urgent).

    Args:
        workers (int, optional): Number of parallel workers (threads). Defaults to twice :py:func:`multiprocessing.cpu_count`.
        priority_levels (int, optional): Number of priority queues. Defaults to 2.
        log_directory (typing.Optional[pathlib.Path], optional): Directory to store log files. Logs are not generated if this is None. Defaults to None.
    """

    class ClosePolicy(enum.Enum):
        """Strategy used to terminate worker threads."""

        JOIN = 0
        """Consume all messages and shutdown threads.

        This should be used to wait for the end of the program normally.
        """

        CANCEL = 1
        """Shutdown threads without consuming buffered messages.

        This should be used to stop thread workers after user-initiated cancellation (CTRL-C).
        """

        KILL = 2
        """Kill threads without consuming buffered messages.

        This should be used after a thread raises an error, to stop the remaning worker threads.
        """

    class Proxy(Manager):
        """Manager interface that can be sent to workers.

        Since :py:class:`ProcessManager` implements a custom message passing system and owns message queues, it cannot be shared between processes. Worker processes require a handle to the manager to send messages and schedule new tasks. However, the handle does not have to be the actual manager, it is merely a means to pass around the two fuctions of its public API. This proxy prentends to be the manager but forwards messages to the actual manager using TCP. See :py:func:`send_bytes` for a description of message encoding.

        Args:
            server_port (int): Port of the manager's TCP server used to send messages between workers and the manager.
        """

        def __init__(self, server_port: int):
            self.server_port = server_port
            self.client: typing.Optional[socket.socket] = None

        def setup(self):
            """Called by each worker to create the TCP connection with the actual manager."""
            self.client = socket.create_connection(
                address=("localhost", self.server_port)
            )

        def schedule(self, task: Task, priority: int = 1):
            logging.debug(f"schedule {task} with priority {priority}")
            assert self.client is not None
            send_message(
                client=self.client,
                type=(128 + priority).to_bytes(1, byteorder="little"),
                message=task,
            )
            receive_type(client=self.client, expected_type=b"s")

        def send_message(self, message: typing.Any):
            assert self.client is not None
            send_message(client=self.client, type=b"m", message=message)
            receive_type(client=self.client, expected_type=b"m")

        def next_task(
            self,
        ) -> typing.Union[None, Task, CloseRequest]:
            """Called by a worker to receive the next task.

            Returns:
                typing.Union[None, Task, CloseRequest]: None if there are no tasks waiting (but more tasks may become available in the future), a :py:class:`Task` instance if the manager returned a task for this worker, and :py:class:`CloseRequest` must shutdown.
            """
            assert self.client is not None
            send_type(client=self.client, type=b"n")
            type, message = receive_message(client=self.client)
            assert type == b"t"
            return message

        def acknowledge_and_next_task(self) -> typing.Union[None, Task, CloseRequest]:
            """Called by a worker to indicate that they completed the current task and are asking for a new one.

            Returns:
                typing.Union[None, Task, CloseRequest]: None if there are no tasks waiting (but more tasks may become available in the future), a :py:class:`Task` instance if the manager returned a task for this worker, and :py:class:`CloseRequest` must shutdown.
            """
            assert self.client is not None
            send_type(client=self.client, type=b"t")
            type, message = receive_message(client=self.client)
            assert type == b"t"
            return message

    def serve(self):
        """Server thread implementation."""
        self.server.serve_forever(poll_interval=0.1)

    @staticmethod
    def target(
        proxy: "ProcessManager.Proxy",
        log_directory: typing.Optional[pathlib.Path],
    ):
        """Worker thread implementation.

        Args:
            proxy (ProcessManager.Proxy): The manager proxy to request tasks, spawn new tasks, and send messages.
            log_directory (typing.Optional[pathlib.Path]): Directory to store log files. Logs are not generated if this is None.
        """
        try:
            if log_directory is not None:
                logging.basicConfig(
                    filename=str(
                        log_directory / f"{multiprocessing.current_process().name}.log"
                    ),
                    encoding="utf-8",
                    level=logging.DEBUG,
                    format="%(asctime)s %(message)s",
                )
            logging.debug("start worker")
            proxy.setup()
            logging.debug("connected to message server")
            with requests.Session() as session:
                active_task: typing.Union[None, Task, CloseRequest] = None
                while True:
                    if active_task is None:
                        active_task = proxy.next_task()
                    if isinstance(active_task, CloseRequest):
                        break
                    if active_task is None:
                        time.sleep(constants.WORKER_POLL_PERIOD)
                        continue
                    try:
                        logging.debug(f"run {active_task}")
                        active_task.run(session=session, manager=proxy)
                    except Exception as exception:
                        logging.debug(exception)
                        proxy.send_message(
                            WorkerException(
                                traceback.TracebackException.from_exception(exception)
                            )
                        )
                    active_task = proxy.acknowledge_and_next_task()

        except KeyboardInterrupt:
            pass
        except ConnectionResetError:
            pass

    def __init__(
        self,
        workers: int = multiprocessing.cpu_count() * 2,
        priority_levels: int = 2,
        log_directory: typing.Optional[pathlib.Path] = None,
    ):
        assert workers > 0
        assert priority_levels > 0 and priority_levels < 128
        self.running = True
        self.message_queue: collections.deque[typing.Any] = collections.deque()
        self.task_queues: tuple[collections.deque[bytes], ...] = tuple(
            collections.deque() for _ in range(0, priority_levels)
        )
        self.tasks_left_lock = threading.Lock()
        self.tasks_left = 0
        self.server = socketserver.ThreadingTCPServer(
            server_address=("localhost", 0),
            RequestHandlerClass=ProcessManager.RequestHandler,
        )
        self.server.daemon_threads = True
        self.server.manager = self  # type: ignore
        logging.debug(
            f"communnication server listening on port {self.server.server_address[1]}"
        )
        self.serve_thread = threading.Thread(target=self.serve, daemon=True)
        self.serve_thread.start()
        self.workers = tuple(
            multiprocessing.Process(
                target=ProcessManager.target,
                daemon=True,
                args=(
                    ProcessManager.Proxy(server_port=self.server.server_address[1]),
                    log_directory,
                ),
            )
            for _ in range(0, workers)
        )
        logging.debug("start workers")
        for worker in self.workers:
            worker.start()

    class RequestHandler(socketserver.BaseRequestHandler):
        """Processes TCP requests for the actual manager (TCP server)."""

        def handle(self):
            """Processes a TCP request.

            See :py:func:`send_bytes` for a description of message encoding.
            """
            logging.debug("start request handler")
            try:
                manager: ProcessManager = self.server.manager  # type: ignore
                while True:
                    type, message = receive_bytes(client=self.request)
                    if type == b"n" or type == b"t":
                        assert message == b""
                        if type == b"t":
                            with manager.tasks_left_lock:
                                manager.tasks_left -= 1
                        if manager.running:
                            task: typing.Optional[bytes] = None
                            for task_queue in manager.task_queues:
                                try:
                                    task = task_queue.popleft()
                                    break
                                except IndexError:
                                    continue
                            if task is None:
                                send_message(
                                    client=self.request, type=b"t", message=None
                                )
                            else:
                                send_bytes(client=self.request, type=b"t", message=task)
                        else:
                            send_message(
                                client=self.request, type=b"t", message=CloseRequest()
                            )
                    elif type == b"m":
                        manager.message_queue.append(pickle.loads(message))
                        send_type(client=self.request, type=b"m")
                    elif type >= b"\x80":
                        manager.task_queues[
                            int.from_bytes(type, byteorder="little") - 128
                        ].append(message)
                        with manager.tasks_left_lock:
                            manager.tasks_left += 1
                        send_type(client=self.request, type=b"s")
                    else:
                        raise Exception(f'unexpected request type "{type}"')
            except ConnectionResetError:
                pass

    def close(self, policy: "ProcessManager.ClosePolicy"):
        """Terminates the manager.

        Depending on  the value of policy, this function will return almost immediately or block until all the tasks complete. See :py:class:`ProcessManager.ClosePolicy` for details.

        Args:
            policy (ProcessManager.ClosePolicy): Termination policy for the manager and its workers.
        """
        logging.debug(f"close manager with policy {policy}")
        if policy == ProcessManager.ClosePolicy.JOIN:
            collections.deque(self.messages(), maxlen=0)
            self.running = False
            for worker in self.workers:
                worker.join()
            self.server.shutdown()
            self.server.server_close()
            self.serve_thread.join()
        elif policy == ProcessManager.ClosePolicy.CANCEL:
            self.running = False
            for worker in self.workers:
                worker.join()
            self.server.shutdown()
            self.server.server_close()
            self.serve_thread.join()
        elif policy == ProcessManager.ClosePolicy.KILL:
            for worker in self.workers:
                worker.kill()
            self.server.shutdown()
            self.server.server_close()
            self.serve_thread.join()

    def __enter__(self):
        """Enables the use of the "with" statement.

        Returns:
            ProcessManager: A process manager context that calls :py:meth:`close` on exit.
        """
        return self

    def __exit__(
        self,
        type: typing.Optional[typing.Type[BaseException]],
        value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ):
        """Enables the use of the "with" statement.

        This function calls :py:meth:`close` with the policy :py:attr:`ProcessManager.ClosePolicy.CANCEL` if there is no active exception (typically caused by a soft cancellation) and with the policy :py:attr:`ProcessManager.ClosePolicy.KILL` if there is an active exception.

        Args:
            type (typing.Optional[typing.Type[BaseException]]): None if the context exits without an exception, and the raised exception's class otherwise.
            value (typing.Optional[BaseException]): None if the context exits without an exception, and the raised exception otherwise.
            traceback (typing.Optional[types.TracebackType]): None if the context exits without an exception, and the raised exception's traceback otherwise.
        """
        logging.debug(f"manager exit with error type {type}")
        if type is None:
            self.close(policy=ProcessManager.ClosePolicy.CANCEL)
        else:
            self.close(policy=ProcessManager.ClosePolicy.KILL)
        return False

    def schedule(self, task: Task, priority: int = 1):
        logging.debug(f"schedule {task} with priority {priority}")
        self.task_queues[priority].append(pickle.dumps(task))
        with self.tasks_left_lock:
            self.tasks_left += 1

    def send_message(self, message: typing.Any):
        self.message_queue.append(message)

    def messages(self) -> typing.Iterable[typing.Any]:
        """Iterates over the messages sent by all workers until all the tasks are complete.

        The thread that iterates over the messages has access to the manager and may use it to schedule new tasks.

        Returns:
            typing.Iterable[typing.Any]: Iterator over messages from all workers.
        """
        while True:
            message: typing.Any = None
            try:
                message = self.message_queue.popleft()
            except IndexError:
                with self.tasks_left_lock:
                    if self.tasks_left == 0:
                        return
                time.sleep(constants.CONSUMER_POLL_PERIOD)
                continue
            yield message



"""Local or remote file."""

from __future__ import annotations

import contextlib
import dataclasses
import functools
import math
import operator
import pathlib
import typing

import requests

from . import constants, decode, exception, remote, task, utilities

if typing.TYPE_CHECKING:
    from . import path_directory

try:
    functools.cached_property
except:
    functools.cached_property = property


class Download(remote.Download):
    """Downloads a remote file.

    This task is never used with a scheduler. Its run function is called by :py:meth:`File.chunks` to recycle the download logic implemented in py:class:`undr.remote.Download`.

    Args:
        path_id (pathlib.PurePosixPath): The resource's unique path id.
        suffix (typing.Optional[str]): Added to the file name while it is being downloaded.
        server (Server): The remote server.
        stream (bool): Whether to download the file in chunks (slightly slower for small files, reduces memory usage for large files).
    """

    def __init__(
        self,
        path_id: pathlib.PurePosixPath,
        suffix: typing.Optional[str],
        server: remote.Server,
        stream: bool,
    ):
        super().__init__(path_id=path_id, suffix=suffix, server=server, stream=stream)
        self.response: typing.Optional[requests.Response] = None

    def on_begin(self, manager: task.Manager) -> int:
        return 0

    def on_response_ready(self, response: requests.Response, manager: task.Manager):
        self.response = response


@dataclasses.dataclass(frozen=True)
class Path:
    """A file or directory in a dataset.

    A path can point to a local resource or represent a remote resource.
    """

    path_root: pathlib.Path
    """Path of the root "datasets" directory used to generate local paths.
    """

    path_id: pathlib.PurePosixPath
    """A POSIX path uniquely identifying the resource (including its dataset).
    """

    own_doi: typing.Optional[str] = dataclasses.field(compare=False, hash=False)
    """This resource's DOI, used by all its children unless they have their own DOI.
    """

    metadata: dict[str, typing.Any] = dataclasses.field(compare=False, hash=False)
    """Any data not strictly required to decode the file (stored in -index.json).
    """

    server: remote.Server = dataclasses.field(repr=False, compare=False, hash=False)
    """The resource's remote server, used to download data if it is not available locally.
    """

    @functools.cached_property
    def local_path(self) -> pathlib.Path:
        """Returns the local file path.

        This function always return a path, even if the local resource does not exist.

        Returns:
            pathlib.Path: The path of the local resource.
        """
        return self.path_root / self.path_id

    def __truediv__(self, other: str) -> "Path":
        """Concatenates this path with a string to create a new path.

        Args:
            other (str): Suffix to append to this path.

        Returns:
            Path: The concatenated result.
        """
        raise NotImplementedError()


@dataclasses.dataclass(frozen=True)
class File(Path):
    """Represents a local or remote file."""

    size: int = dataclasses.field(compare=False, hash=False)
    """The decompressed file size in bytes.
    """

    hash: str = dataclasses.field(compare=False, hash=False)
    """The decompressed file hash (SHA3-224).
    """

    compressions: tuple[decode.Compression, ...] = dataclasses.field(
        repr=False, compare=False, hash=False
    )
    """List of compressions available on the server.
    """

    session: typing.Optional[requests.Session] = dataclasses.field(
        repr=False, compare=False, hash=False
    )
    """An open session that can be used to download resources.
    """

    manager: task.Manager = dataclasses.field(repr=False, compare=False, hash=False)
    """Can be called to schedule new tasks and report updates.
    """

    @functools.cached_property
    def best_compression(self) -> decode.Compression:
        """Returns the best compression supported by the remote server for this file.

        Best is defined here as "smallest encoded size".

        Returns:
            decode.Compression: Compression format that yields the smallest version of this file.
        """
        return min(self.compressions, key=operator.attrgetter("size"))

    @functools.cached_property
    def word_size(self) -> int:
        """The size of an entry in this file, in bytes.

        This can be used to ensure that entries (events, frames...) are not split while reading.
        A decoded file's size in bytes must be a multiple of the value returned by this function.

        Returns:
            int: Number of bytes used by each entry.
        """
        return 1

    @staticmethod
    def attributes_from_dict(
        data: dict[str, typing.Any], parent: "path_directory.Directory"
    ) -> dict[str, typing.Any]:
        """Converts -index.json data to a dict of this class's arguments.

        The returned dict can be used to initialise an instance of this class.

        Args:
            data (dict[str, typing.Any]): Parsed JSON data.
            parent (path_directory.Directory): The file's parent directory.

        Returns:
            dict[str, typing.Any]: Data that can be used to initialize this class.
        """
        return {
            "path_root": parent.path_root,
            "path_id": parent.path_id / data["name"],
            "own_doi": data["doi"] if "doi" in data else None,
            "metadata": data["metadata"],
            "server": parent.server,
            "size": data["size"],
            "hash": data["hash"],
            "compressions": tuple(
                decode.compression_from_dict(
                    data=compression,
                    base_size=data["size"],
                    base_hash=data["hash"],
                )
                for compression in data["compressions"]
            ),
            "session": None,
            "manager": task.NullManager(),
        }

    @classmethod
    def from_dict(cls, data: dict[str, typing.Any], parent: "path_directory.Directory"):
        """Conerts -index.json data to an instance of this class.

        Args:
            data (dict[str, typing.Any]): Parsed JSON data.
            parent (path_directory.Directory): The file's parent directory.

        Returns:
            File: The file represented by the given data.
        """
        return cls(**File.attributes_from_dict(data=data, parent=parent))

    def __truediv__(self, other: str) -> Path:
        raise Exception(f'the file path "{self.path_id}" cannot be extended')

    def _chunks(self, word_size: int) -> typing.Iterable[bytes]:
        """Returns an iterator over the file's decompressed bytes.

        Users should prefer :py:func:`chunks` since files know their word size.

        Args:
            word_size (int): size of an entry (events, frames...) in the file.

        Raises:
            decode.RemainingBytesError: if the total file size is not a multiple of `word_size`.
            Exception: if the hash is incorrect.

        Returns:
            typing.Iterable[bytes]: Iterator over the file's decompressed bytes.
        """
        assert word_size > 0
        if self.local_path.is_file():
            hash = utilities.new_hash()
            chunk_size = math.ceil(65536 / word_size) * word_size
            with open(self.local_path, "rb") as file_object:
                while True:
                    chunk = file_object.read(chunk_size)
                    if len(chunk) == 0:
                        break
                    if len(chunk) % word_size != 0:
                        raise decode.RemainingBytesError(
                            word_size=word_size, buffer=chunk
                        )
                    yield chunk
                    hash.update(chunk)
                    self.manager.send_message(
                        decode.Progress(
                            path_id=self.path_id,
                            initial_bytes=0,
                            current_bytes=len(chunk),
                            final_bytes=len(chunk),
                            complete=False,
                        )
                    )
            digest = hash.hexdigest()
            if digest != self.hash:
                raise exception.HashMismatch(self.path_id, self.hash, digest)
            self.manager.send_message(
                decode.Progress(
                    path_id=self.path_id,
                    initial_bytes=0,
                    current_bytes=0,
                    final_bytes=0,
                    complete=True,
                )
            )
        elif utilities.path_with_suffix(
            self.local_path, self.best_compression.suffix
        ).is_file():
            hash = utilities.new_hash()
            with open(
                utilities.path_with_suffix(
                    self.local_path, self.best_compression.suffix
                ),
                "rb",
            ) as compressed_file:
                decoder = self.best_compression.decoder(self.word_size)
                while True:
                    encoded_bytes = compressed_file.read(constants.CHUNK_SIZE)
                    if len(encoded_bytes) == 0:
                        break
                    decoded_bytes = decoder.decompress(encoded_bytes)
                    yield decoded_bytes
                    hash.update(decoded_bytes)
                    self.manager.send_message(
                        decode.Progress(
                            path_id=self.path_id,
                            initial_bytes=0,
                            current_bytes=len(decoded_bytes),
                            final_bytes=len(decoded_bytes),
                            complete=False,
                        )
                    )
                decoded_bytes, remaining_bytes = decoder.finish()
                if len(decoded_bytes) > 0:
                    yield decoded_bytes
                    hash.update(decoded_bytes)
                    self.manager.send_message(
                        decode.Progress(
                            path_id=self.path_id,
                            initial_bytes=0,
                            current_bytes=len(decoded_bytes),
                            final_bytes=len(decoded_bytes),
                            complete=False,
                        )
                    )
                if len(remaining_bytes) > 0:
                    raise decode.RemainingBytesError(word_size, remaining_bytes)
            digest = hash.hexdigest()
            if digest != self.hash:
                raise exception.HashMismatch(self.path_id, self.hash, digest)
            self.manager.send_message(
                decode.Progress(
                    path_id=self.path_id,
                    initial_bytes=0,
                    current_bytes=0,
                    final_bytes=0,
                    complete=True,
                )
            )
        else:
            download_hash = utilities.new_hash()
            decode_hash = utilities.new_hash()
            with (
                requests.Session() if self.session is None else contextlib.nullcontext()
            ) as local_session:
                if self.session is None:
                    self.attach_session(local_session)
                    assert self.session is not None
                decoder = self.best_compression.decoder(self.word_size)
                # this task is created to re-use download logic but it is never scheduled
                # we call run directy below
                download = Download(
                    path_id=self.path_id,
                    suffix=self.best_compression.suffix,
                    server=self.server,
                    stream=self.size
                    >= constants.CHUNK_SIZE * constants.STREAM_CHUNK_THRESHOLD,
                )
                download.run(session=self.session, manager=self.manager)
                assert download.response is not None
                for encoded_bytes in download.response.iter_content(
                    constants.CHUNK_SIZE
                ):
                    download_hash.update(encoded_bytes)
                    self.manager.send_message(
                        remote.Progress(
                            path_id=self.path_id,
                            initial_bytes=0,
                            current_bytes=len(encoded_bytes),
                            final_bytes=len(encoded_bytes),
                            complete=False,
                        )
                    )
                    decoded_bytes = decoder.decompress(encoded_bytes)
                    yield decoded_bytes
                    decode_hash.update(decoded_bytes)
                    self.manager.send_message(
                        decode.Progress(
                            path_id=self.path_id,
                            initial_bytes=0,
                            current_bytes=len(decoded_bytes),
                            final_bytes=len(decoded_bytes),
                            complete=False,
                        )
                    )
                download.response.close()
            download_digest = download_hash.hexdigest()
            if download_digest != self.best_compression.hash:
                raise Exception(
                    f'bad download hash for "{self.path_id}" (expected "{self.best_compression.hash}", got "{download_digest}")'
                )
            self.manager.send_message(
                remote.Progress(
                    path_id=self.path_id,
                    initial_bytes=0,
                    current_bytes=0,
                    final_bytes=0,
                    complete=True,
                )
            )
            decoded_bytes, remaining_bytes = decoder.finish()
            if len(decoded_bytes) > 0:
                yield decoded_bytes
                decode_hash.update(decoded_bytes)
                self.manager.send_message(
                    decode.Progress(
                        path_id=self.path_id,
                        initial_bytes=0,
                        current_bytes=len(decoded_bytes),
                        final_bytes=len(decoded_bytes),
                        complete=False,
                    )
                )
            if len(remaining_bytes) > 0:
                raise decode.RemainingBytesError(word_size, remaining_bytes)
            decode_digest = decode_hash.hexdigest()
            if decode_digest != self.hash:
                raise Exception(
                    f'bad download hash for "{self.path_id}" (expected "{self.hash}", got "{decode_digest}")'
                )
            self.manager.send_message(
                decode.Progress(
                    path_id=self.path_id,
                    initial_bytes=0,
                    current_bytes=0,
                    final_bytes=0,
                    complete=True,
                )
            )

    def chunks(self) -> typing.Iterable[bytes]:
        """Returns an iterator over the file's decompressed bytes.

        Returns:
            typing.Iterable[bytes]: Iterator over the decompressed file's bytes. The size of the chunks may vary.
        """
        yield from self._chunks(word_size=1)

    def attach_session(self, session: typing.Optional[requests.Session]):
        """Binds a session to this file.

        The session is used for all subsequent downloads.

        Args:
            session (typing.Optional[requests.Session]): An open session to use for downloads.
        """
        self.__dict__["session"] = session

    def attach_manager(self, manager: typing.Optional[task.Manager]):
        """Binds a manager to this file.

        The file sends all subsequent updates (download and processing) to the manager.

        Args:
            manager (typing.Optional[task.Manager]): The manager to use to keep track of progress.
        """
        self.__dict__["manager"] = manager



"""Constants used throughout the codebase."""

CHUNK_SIZE: int = 65536
"""Buffer size in bytes for file reads."""

CONSUMER_POLL_PERIOD: float = 0.1
"""Sleep duration for msssage readers."""

DECOMPRESS_SUFFIX: str = ".decompress"
"""Suffix indicating that a file is being decompressed."""

DEFAULT_TIMEOUT: float = 60.0
"""Server timeout, can be overriden in settings TOML files."""

DOWNLOAD_SUFFIX: str = ".download"
"""Suffix indicating that a file is being downloaded."""

LRU_CACHE_MAXSIZE: int = 128
"""Number of index files cached by the load function."""

SPEED_SAMPLES: int = 30
"""Number of samples used to smooth the speed measurement (sliding window)."""

STREAM_CHUNK_THRESHOLD: int = 64
"""Below this number, files are download in one chunk instead of several to boost performance."""

WORKER_POLL_PERIOD: float = 0.02
"""Sleep duration for wworkers when the task queue is empty."""



"""User-facing API to manipulate configuration files and trigger actions (download, decompress...)."""

from __future__ import annotations

import contextlib
import dataclasses
import logging
import multiprocessing
import os
import pathlib
import typing

import requests

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

from . import (bibtex, constants, display, formats, install_mode, json_index,
               json_index_tasks, path, path_directory, persist, remote, task,
               utilities)

schema = utilities.load_schema("undr_schema.json")
"""JSON schema for TOML settings files."""


class InstallSelector(json_index_tasks.Selector):
    """Selector for a standard installation, maps install modes to actions.

    Raises:
        ValueError: if mode is not :py:attr:`undr.install_mode.Mode.REMOTE`, :py:attr:`undr.install_mode.Mode.LOCAL`, or :py:attr:`undr.install_mode.Mode.RAW`.
    """

    def __init__(self, mode: install_mode.Mode):
        self.scan = False
        if mode == install_mode.Mode.REMOTE:
            self.cached_action = json_index_tasks.Selector.Action.IGNORE
        elif mode == install_mode.Mode.LOCAL:
            self.cached_action = json_index_tasks.Selector.Action.DOWNLOAD
            self.scan = True
        elif mode == install_mode.Mode.RAW:
            self.cached_action = json_index_tasks.Selector.Action.DECOMPRESS
            self.scan = True
        else:
            raise ValueError(f'unexpected mode "{mode}"')

    def action(self, file: path.File) -> json_index_tasks.Selector.Action:
        return self.cached_action

    def scan_filesystem(self, directory: path_directory.Directory):
        return self.scan


class DoiSelector(json_index_tasks.Selector):
    """Selector for a DOI download."""

    def action(self, file: path.File) -> json_index_tasks.Selector.Action:
        return json_index_tasks.Selector.Action.DOI

    def scan_filesystem(self, directory: path_directory.Directory):
        return False


@dataclasses.dataclass
class DatasetSettings:
    """A dataset entry in a TOML settings file."""

    name: str
    """The dataset's name, used to name the local directory.
    """

    url: str
    """The dataset's base URL.
    """

    mode: install_mode.Mode
    """The installation mode.
    """

    timeout: typing.Optional[float]
    """Request timeout in seconds.
    """


@dataclasses.dataclass
class IndexStatus:
    """Keeps track of the indexing progress for a dataset."""

    dataset_settings: DatasetSettings
    """User-specified dataset settings.
    """

    current_index_files: int
    """Number of index files parsed.

    The dataset has been indexed if :py:attr:`current_index_files` and :py:attr:`final_index_files` are equal.
    """

    final_index_files: int
    """Total number of index files.
    """

    server: remote.Server
    """The remote server for this dataset.
    """

    selector: json_index_tasks.Selector
    """Selector to choose actions while indexing.
    """

    downloaded_and_processed: bool
    """Whether the dataset has been fully downloaded and processed.
    """

    def push(self, message: typing.Any) -> tuple[bool, typing.Optional["IndexStatus"]]:
        """Updates the status based on the message.

        Ignores messages that are not :py:class:`undr.json_index_tasks.IndexLoaded` or :py:class:`undr.json_index_tasks.DirectoryScanned`.

        Returns:
            tuple[bool, typing.Optional["IndexStatus"]]: Whether the dataset has been fully indexed and self, if self was updated.
        """
        if isinstance(message, json_index_tasks.IndexLoaded):
            self.final_index_files += message.children
            return (False, self)
        if isinstance(message, json_index_tasks.DirectoryScanned):
            self.current_index_files += 1
            if (
                message.download_bytes.initial != message.download_bytes.final
                or message.process_bytes.initial != message.process_bytes.final
            ):
                self.downloaded_and_processed = False
            return (self.current_index_files == self.final_index_files, self)
        return False, None


@dataclasses.dataclass
class IndexesStatuses:
    """Maps dataset names to index statuses."""

    name_to_status: dict[str, IndexStatus]
    """Inner dict.
    """

    def push(self, message: typing.Any) -> tuple[bool, typing.Optional[IndexStatus]]:
        """Processes relevant messages.

        This function updates the indexing status and returns it
        if message is a :py:class:`undr.json_index_tasks.IndexLoaded` or :py:class:`undr.json_index_tasks.DirectoryScanned` object.
        If the message was the last indexing message for this dataset, the first argument is True.
        """
        if isinstance(
            message, (json_index_tasks.IndexLoaded, json_index_tasks.DirectoryScanned)
        ):
            return self.name_to_status[message.path_id.parts[0]].push(message=message)
        return False, None


@dataclasses.dataclass
class MapMessage:
    """A message generated by :py:class:`MapSelector`."""

    payload: typing.Any
    """Payload attached to this message.

    The payload may be any object type. The user is reponsible for checking message types.
    """


class MapSelector(json_index_tasks.Selector):
    """Applies a user-provided function to each file.

    Args:
        enabled_types (set[typing.Any]): The file types (a class in :py:mod:`undr.formats`) to process.
        store (typing.Optional[persist.ReadOnlyStore]): A store to check for readily processed files.
    """

    def __init__(
        self,
        enabled_types: set[typing.Any],
        store: typing.Optional[persist.ReadOnlyStore],
    ):
        self.enabled_types = enabled_types
        self.store = store

    def action(self, file: path.File):
        if file.__class__ in self.enabled_types:
            if self.store is not None and str(file.path_id) in self.store:
                return json_index_tasks.Selector.Action.SKIP
            return json_index_tasks.Selector.Action.PROCESS
        return json_index_tasks.Selector.Action.IGNORE

    def scan_filesystem(self, directory: path_directory.Directory):
        return len(self.enabled_types) > 0


class MapProcessFile(json_index_tasks.ProcessFile):
    """Uses a switch to process files and wraps messages into :py:class:`MapMessage`.

    Args:
        file (path.File): The file to process.
        switch (formats.Switch): Switch that maps file types to actions.
    """

    def __init__(self, file: path.File, switch: formats.Switch):
        super().__init__(file=file)
        self.switch = switch

    def run(self, session: requests.Session, manager: task.Manager):
        self.file.attach_session(session)
        self.file.attach_manager(manager)
        self.switch.handle_file(
            file=self.file,
            send_message=lambda message: manager.send_message(
                MapMessage(payload=message)
            ),
        )
        manager.send_message(persist.Progress(path_id=self.file.path_id))


@dataclasses.dataclass
class Configuration:
    """Represents a dataset configuration (TOML)."""

    directory: pathlib.Path
    """Local path of the root datasets directory (usually called *datasets*).
    """

    name_to_dataset_settings: dict[str, DatasetSettings]
    """Maps dataset names to their parameters.
    """

    def dataset(self, name: str) -> path_directory.Directory:
        """Returns the dataset with the given name.

        Args:
            name (str): The dataset name.

        Raises:
            ValueError: if the dataset exists but is disabled.

        Returns:
            path_directory.Directory: The dataset's root directory.
        """
        dataset_settings = self.name_to_dataset_settings[name]
        if dataset_settings.mode == install_mode.Mode.DISABLED:
            raise ValueError(f'"{name}" is disabled')
        return path_directory.Directory(
            path_root=self.directory,
            path_id=pathlib.PurePosixPath(dataset_settings.name),
            own_doi=None,
            metadata={},
            server=remote.Server(
                url=dataset_settings.url,
                timeout=constants.DEFAULT_TIMEOUT
                if dataset_settings.timeout is None
                else dataset_settings.timeout,
            ),
            doi_and_metadata_loaded=False,
        )

    def iter(self, recursive: bool = False) -> typing.Iterable[path.Path]:
        """Iterates the files in the dataset.

        Args:
            recursive (bool, optional): Whether to recursively search child directories. Defaults to False.

        Returns:
            typing.Iterable[path.Path]: Iterator over the child paths. If `recursive` is false, the iterator yields the direct children (files and directories) of the root dataset directory. If `recursive` is true, the iterator yields all the children (files and directories) of the dataset.
        """
        for dataset_settings in self.enabled_datasets_settings():
            if dataset_settings.mode != install_mode.Mode.DISABLED:
                directory = path_directory.Directory(
                    path_root=self.directory,
                    path_id=pathlib.PurePosixPath(dataset_settings.name),
                    own_doi=None,
                    metadata={},
                    server=remote.Server(
                        url=dataset_settings.url,
                        timeout=constants.DEFAULT_TIMEOUT
                        if dataset_settings.timeout is None
                        else dataset_settings.timeout,
                    ),
                    doi_and_metadata_loaded=False,
                )
                if recursive:
                    yield from directory.iter(recursive=True)
                else:
                    yield directory

    def enabled_datasets_settings(self) -> list[DatasetSettings]:
        """The settings of enabled datasets.

        The list always contains at least one item (the function otherwise raises an error).

        Raises:
            RuntimeError: if all the datasets are disabled or there are no datasets.

        Returns:
            list[DatasetSettings]: The settings of the datasets that are enabled, in the same order as the configuration file.
        """
        result = [
            dataset_settings
            for dataset_settings in self.name_to_dataset_settings.values()
            if dataset_settings.mode != install_mode.Mode.DISABLED
        ]
        if len(result) == 0:
            raise RuntimeError(
                "the configuration is empty or all the datasets are disabled"
            )
        return result

    def display(
        self,
        download_tag: display.Tag = display.Tag(label="download", icon="↓"),
        process_tag: display.Tag = display.Tag(label="process", icon="⚛"),
    ) -> display.Display:
        """Returns a display that shows download and process progress for enabled datasets.

        Args:
            download_tag (display.Tag, optional): Label and icon for download. Defaults to display.Tag(label="download", icon="↓").
            process_tag (display.Tag, optional): Label and icon for process. Defaults to display.Tag(label="process", icon="⚛").

        Returns:
            display.Display: Controller for the display thread.
        """
        return display.Display(
            statuses=[
                display.Status.from_path_id_and_mode(
                    path_id=pathlib.PurePosixPath(dataset_settings.name),
                    # display both progress bars by default (RAW dataset mode)
                    dataset_mode=install_mode.Mode.RAW,
                )
                for dataset_settings in self.enabled_datasets_settings()
            ],
            output_interval=constants.CONSUMER_POLL_PERIOD,
            download_speed_samples=constants.SPEED_SAMPLES,
            process_speed_samples=constants.SPEED_SAMPLES,
            download_tag=download_tag,
            process_tag=process_tag,
        )

    def indexes_statuses(self, selector: json_index_tasks.Selector) -> IndexesStatuses:
        """Builds an indexing report for enabled datasets.

        Args:
            selector (json_index_tasks.Selector): The selector used to index the dataset.

        Returns:
            IndexesStatuses: Index status for enaabled datasets, in the same order as the configuration file.
        """
        return IndexesStatuses(
            name_to_status={
                dataset_settings.name: IndexStatus(
                    dataset_settings=dataset_settings,
                    current_index_files=0,
                    final_index_files=1,
                    server=remote.Server(
                        url=dataset_settings.url,
                        timeout=constants.DEFAULT_TIMEOUT
                        if dataset_settings.timeout is None
                        else dataset_settings.timeout,
                    ),
                    selector=selector,
                    downloaded_and_processed=True,
                )
                for dataset_settings in self.enabled_datasets_settings()
            }
        )

    def install(
        self,
        show_display: bool,
        workers: int,
        force: bool,
        log_directory: typing.Optional[pathlib.Path],
    ):
        """Downloads index files and data files and decompresses data files.

        The action (index only, download, download and decompress) may be different for each dataset and is controlled by :py:class:`undr.install_mode.Mode`.

        Args:
            show_display (bool): Whether to show progress in the terminal.
            workers (int): Number of parallel workers (threads).
            force (bool): Whether to re-download resources even if they are already present locally.
            log_directory (typing.Optional[pathlib.Path]): Directory to store log files. Logs are not generated if this is None.

        Raises:
            task.WorkerException: if a worker raises an error.
        """
        with (
            display.Display(
                statuses=[
                    display.Status.from_path_id_and_mode(
                        path_id=pathlib.PurePosixPath(dataset_settings.name),
                        dataset_mode=dataset_settings.mode,
                    )
                    for dataset_settings in self.enabled_datasets_settings()
                ],
                output_interval=constants.CONSUMER_POLL_PERIOD,
                download_speed_samples=constants.SPEED_SAMPLES,
                process_speed_samples=constants.SPEED_SAMPLES,
                download_tag=display.Tag(label="download", icon="↓"),
                process_tag=display.Tag(label="decompress", icon="↔"),
            )
            if show_display
            else contextlib.nullcontext()
        ) as progress_display, task.ProcessManager(
            workers=workers, priority_levels=2, log_directory=log_directory
        ) as manager:
            indexes_statuses = IndexesStatuses(
                name_to_status={
                    dataset_settings.name: IndexStatus(
                        dataset_settings=dataset_settings,
                        current_index_files=0,
                        final_index_files=1,
                        server=remote.Server(
                            url=dataset_settings.url,
                            timeout=constants.DEFAULT_TIMEOUT
                            if dataset_settings.timeout is None
                            else dataset_settings.timeout,
                        ),
                        selector=InstallSelector(dataset_settings.mode),
                        downloaded_and_processed=True,
                    )
                    for dataset_settings in self.enabled_datasets_settings()
                }
            )
            for dataset_settings in self.enabled_datasets_settings():
                manager.schedule(
                    task=json_index_tasks.Index(
                        path_root=self.directory,
                        path_id=pathlib.PurePosixPath(dataset_settings.name),
                        server=indexes_statuses.name_to_status[
                            dataset_settings.name
                        ].server,
                        selector=indexes_statuses.name_to_status[
                            dataset_settings.name
                        ].selector,
                        priority=0,
                        force=force,
                        directory_doi=False,
                    ),
                    priority=0,
                )
            for message in manager.messages():
                if isinstance(message, task.WorkerException):
                    raise message
                if progress_display is not None:
                    progress_display.push(message=message)
                indexing_complete, status = indexes_statuses.push(message=message)
                if (
                    indexing_complete
                    and status is not None
                    and not status.downloaded_and_processed
                ):
                    manager.schedule(
                        task=json_index_tasks.InstallFilesRecursive(
                            path_root=self.directory,
                            path_id=pathlib.PurePosixPath(status.dataset_settings.name),
                            server=status.server,
                            selector=status.selector,
                            priority=1,
                            force=force,
                        ),
                        priority=1,
                    )

    def bibtex(
        self,
        show_display: bool,
        workers: int,
        force: bool,
        bibtex_timeout: float,
        log_directory: typing.Optional[pathlib.Path],
    ) -> str:
        """Downloads index files and BibTeX references for enabled datasets.

        Args:
            show_display (bool): Whether to show progress in the terminal.
            workers (int): Number of parallel workers (threads).
            force (bool): Whether to re-download resources even if they are already present locally.
            bibtex_timeout (float): Timeout for requests to https://dx.doi.org/.
            log_directory (typing.Optional[pathlib.Path]): Directory to store log files. Logs are not generated if this is None.

        Raises:
            task.WorkerException: if a worker raises an error.

        Returns:
            str: BibTeX references as a string.
        """
        with (
            display.Display(
                statuses=[
                    display.Status.from_path_id_and_mode(
                        path_id=pathlib.PurePosixPath(dataset_settings.name),
                        dataset_mode=install_mode.Mode.REMOTE,
                    )
                    for dataset_settings in self.enabled_datasets_settings()
                ],
                output_interval=constants.CONSUMER_POLL_PERIOD,
                download_speed_samples=constants.SPEED_SAMPLES,
                process_speed_samples=constants.SPEED_SAMPLES,
                download_tag=display.Tag(label="download", icon="↓"),
                process_tag=display.Tag(label="", icon=""),
            )
            if show_display
            else contextlib.nullcontext()
        ) as progress_display, task.ProcessManager(
            workers=workers, priority_levels=2, log_directory=log_directory
        ) as manager:
            selector = DoiSelector()
            for dataset_settings in self.enabled_datasets_settings():
                manager.schedule(
                    task=json_index_tasks.Index(
                        path_root=self.directory,
                        path_id=pathlib.PurePosixPath(dataset_settings.name),
                        server=remote.Server(
                            url=dataset_settings.url,
                            timeout=constants.DEFAULT_TIMEOUT
                            if dataset_settings.timeout is None
                            else dataset_settings.timeout,
                        ),
                        selector=selector,
                        priority=0,
                        force=force,
                        directory_doi=False,
                    ),
                    priority=0,
                )
            doi_to_path_ids_and_bibtex: dict[
                str, tuple[list[pathlib.PurePosixPath], str]
            ] = {}
            for message in manager.messages():
                if isinstance(message, task.WorkerException):
                    raise message
                if progress_display is not None:
                    progress_display.push(message)
                if isinstance(message, json_index_tasks.Doi):
                    if message.value in doi_to_path_ids_and_bibtex:
                        doi_to_path_ids_and_bibtex[message.value][0].append(
                            message.path_id
                        )
                    else:
                        try:
                            doi_to_path_ids_and_bibtex[message.value] = (
                                [message.path_id],
                                bibtex.from_doi(
                                    doi=message.value,
                                    pretty=True,
                                    timeout=bibtex_timeout,
                                ),
                            )
                        except (
                            requests.HTTPError,
                            requests.ConnectionError,
                        ) as exception:
                            doi_to_path_ids_and_bibtex[message.value] = (
                                [message.path_id],
                                f"% downloading application/x-bibtex data from https://dx.doi.org/{message.value} failed, {exception}\n",
                            )
        path_ids_and_bibtexs = [
            (sorted(path_ids), bibtex)
            for _, (path_ids, bibtex) in doi_to_path_ids_and_bibtex.items()
        ]
        path_ids_and_bibtexs.sort(
            key=lambda path_ids_and_bibtex: path_ids_and_bibtex[0][0]
        )
        result = ""
        for path_ids, bibtex_content in path_ids_and_bibtexs:
            if len(result) > 0:
                result += "\n"
            if len(path_ids) > 5:
                result += (
                    f"% {', '.join(str(path_id) for path_id in path_ids[:3])}"
                    + f", ... ({len(path_ids) - 4} more), {path_ids[-1]}\n"
                )
            else:
                result += f"% {', '.join(str(path_id) for path_id in path_ids)}\n"
            result += bibtex_content
        return result

    def map(
        self,
        switch: formats.Switch,
        store: typing.Optional[persist.Store] = None,
        show_display: bool = True,
        workers: int = multiprocessing.cpu_count() * 2,
        log_directory: typing.Optional[pathlib.Path] = None,
    ) -> typing.Iterable[typing.Any]:
        """Applies a function to eacch file in a dataset.

        Args:
            switch (formats.Switch): Specifies the action to perform on each file type.
            store (typing.Optional[persist.Store], optional): Saves progress, makes it possible to resume interrupted processing. Defaults to None.
            show_display (bool, optional): Whether to show progress in the terminal. Defaults to True.
            workers (int, optional): Number of parallel workers (threads). Defaults to twice :py:func:`multiprocessing.cpu_count`.
            log_directory (typing.Optional[pathlib.Path], optional): Directory to store log files. Logs are not generated if this is None. Defaults to None.

        Raises:
            task.WorkerException: if a worker raises an error.

        Returns:
            typing.Iterable[typing.Any]: Iterator over the non-error messages generated by the workers.
        """
        selector = MapSelector(
            enabled_types=switch.enabled_types(),
            store=None if store is None else persist.ReadOnlyStore(path=store.path),
        )
        with (
            self.display() if show_display else contextlib.nullcontext()
        ) as progress_display, task.ProcessManager(
            workers=workers, priority_levels=2, log_directory=log_directory
        ) as manager:
            indexes_statuses = self.indexes_statuses(selector=selector)
            for dataset_settings in self.enabled_datasets_settings():
                manager.schedule(
                    task=json_index_tasks.Index(
                        path_root=self.directory,
                        path_id=pathlib.PurePosixPath(dataset_settings.name),
                        server=indexes_statuses.name_to_status[
                            dataset_settings.name
                        ].server,
                        selector=indexes_statuses.name_to_status[
                            dataset_settings.name
                        ].selector,
                        priority=0,
                        force=False,
                        directory_doi=False,
                    ),
                    priority=0,
                )
            for message in manager.messages():
                logging.debug(f"{message=}")
                if isinstance(message, task.WorkerException):
                    raise message
                if store is not None and isinstance(message, persist.Progress):
                    store.add(str(message.path_id))
                if progress_display is not None:
                    progress_display.push(message=message)
                indexing_complete, status = indexes_statuses.push(message=message)
                logging.debug(f"{indexing_complete=}, {status=}")
                if (
                    indexing_complete
                    and status is not None
                    and not status.downloaded_and_processed
                ):
                    manager.schedule(
                        task=json_index_tasks.ProcessFilesRecursive(
                            path_root=self.directory,
                            path_id=pathlib.PurePosixPath(status.dataset_settings.name),
                            server=indexes_statuses.name_to_status[
                                status.dataset_settings.name
                            ].server,
                            process_file_class=MapProcessFile,
                            process_file_args=(),
                            process_file_kwargs={"switch": switch},
                            selector=indexes_statuses.name_to_status[
                                status.dataset_settings.name
                            ].selector,
                            priority=1,
                        ),
                        priority=1,
                    )
                if isinstance(message, MapMessage):
                    yield message.payload

    def mktree(
        self,
        root: typing.Union[str, os.PathLike],
        parents: bool = False,
        exist_ok: bool = False,
    ):
        """Creates a copy of the datasets' file hierarchy without the index or data files.

        This function can be combined with :py:func:`map` to implement a map-reduce algorithm over entire datasets.
            a. Use ``mktree`` to create a empty copy of the file hierarchy.
            b. Use :py:meth:`Configuration.map` to create a result file in the new hierarchy for each data file in the originall hierarchy (for instance, a file that contains a measure algorithm's performance as a single number).
            c. Collect the results ("reduce") by reading the result files in the new hierarchy.

        This approach has several benefits. The most expensive step b. runs in parallell and can be interrupted and resumed. Result files are stored in a different directory and can easily be deleted without altering the original data. The new file hierarchy prevents name clashes as long as result files are named after data files, and workers do not need to worry about directory existence since ``mktree`` runs first.

        Args:
            root (typing.Union[str, os.PathLike]): Directory where the new file hierarchy is created.
            parents (bool, optional): Whether to create the parents of the new directory, if they do not exist. Defaults to False.
            exist_ok (bool, optional): Whether to silence exeptions if the root directory already exists. Defaults to False.
        """
        root = pathlib.Path(root).resolve()
        root.mkdir(parents=parents, exist_ok=exist_ok)

        def mkdir_recursive(
            path_id: pathlib.PurePosixPath,
            read_root: pathlib.Path,
            write_root: pathlib.Path,
            exist_ok: bool,
        ):
            (write_root / path_id).mkdir(exist_ok=exist_ok)
            for child_directory_name in json_index.load(
                read_root / path_id / "-index.json"
            )["directories"]:
                mkdir_recursive(
                    path_id=path_id / child_directory_name,
                    read_root=read_root,
                    write_root=write_root,
                    exist_ok=exist_ok,
                )

        for dataset_settings in self.enabled_datasets_settings():
            mkdir_recursive(
                path_id=pathlib.PurePosixPath(dataset_settings.name),
                read_root=self.directory,
                write_root=root,
                exist_ok=exist_ok,
            )


def configuration_from_path(path: typing.Union[str, os.PathLike]) -> Configuration:
    """Reads the configuration (TOML) with the given path.

    Args:
        path (typing.Union[str, os.PathLike]): Configuration file path.

    Raises:
        RuntimeError: if two datasets have the same name in the configuraation.

    Returns:
        Configuration: the parsed TOML configuration.
    """
    path = pathlib.Path(path).resolve()
    with open(path, "rb") as configuration_file:
        configuration = tomllib.load(configuration_file)
    schema.validate(configuration)
    names = set()
    for dataset in configuration["datasets"]:
        if dataset["name"] in names:
            raise RuntimeError(
                f"two datasets share the same name \"{dataset['name']}\" in \"{path}\""
            )
        names.add(dataset["name"])
    directory = pathlib.Path(configuration["directory"])
    if not directory.is_absolute():
        directory = path.parent / directory
    directory.mkdir(exist_ok=True, parents=True)
    return Configuration(
        directory=directory,
        name_to_dataset_settings={
            dataset["name"]: DatasetSettings(
                name=dataset["name"],
                url=dataset["url"],
                mode=install_mode.Mode(dataset["mode"]),
                timeout=dataset["timeout"] if "timeout" in dataset else None,
            )
            for dataset in configuration["datasets"]
        },
    )



"""Implementation of datasets conformity tests (file structure, hashes, event coordinates...)."""

from __future__ import annotations

import collections
import dataclasses
import itertools
import json
import pathlib
import typing

import numpy
import requests

from . import (decode, formats, json_index, json_index_tasks, path,
               path_directory, remote, task, utilities)


@dataclasses.dataclass
class Error:
    """Reports invariant violations while checking data files."""

    path_id: pathlib.PurePosixPath
    """Identifier of the problematic resource.
    """

    message: str
    """Description of the problem.
    """


def handle_aps(file: formats.ApsFile, send_message: formats.SendMessage):
    """Checks the invariants of an APS file.

    Args:
        file (formats.ApsFile): The file to check.
        send_message (formats.SendMessage): Callback channel for errors.
    """
    non_monotonic_ts = 0
    size_mismatches = 0
    empty = True
    try:
        previous_t = 0
        for frames in file.packets():
            empty = False
            if len(frames) > 0 and frames["t"][0] < previous_t:
                non_monotonic_ts += 1
                previous_t = frames["t"][-1]
            non_monotonic_ts += numpy.count_nonzero(
                numpy.diff(frames["t"].astype("<i8")) < 0
            )
            size_mismatches = numpy.count_nonzero(
                numpy.logical_or(
                    frames["width"] != file.width, frames["height"] != file.height
                )
            )
    except decode.RemainingBytesError as error:
        empty = False
        send_message(
            Error(path_id=file.path_id, message=f"{len(error.buffer)} extra bytes")
        )
    if empty:
        send_message(Error(path_id=file.path_id, message="no data found"))
    if non_monotonic_ts > 0:
        send_message(
            Error(
                path_id=file.path_id,
                message="{} non-monotonic timestamp{}".format(
                    non_monotonic_ts, "s" if non_monotonic_ts > 1 else ""
                ),
            )
        )
    if size_mismatches > 0:
        send_message(
            Error(
                path_id=file.path_id,
                message="{} size mismatch{}".format(
                    size_mismatches, "es" if size_mismatches > 1 else ""
                ),
            )
        )


def handle_dvs(file: formats.DvsFile, send_message: formats.SendMessage):
    """Checks the invariants of a DVS file.

    Args:
        file (formats.DvsFile): The file to check.
        send_message (formats.SendMessage): Callback channel for errors.
    """
    non_monotonic_ts = 0
    out_of_bounds_count = 0
    empty = True
    try:
        previous_t = 0
        for events in file.packets():
            empty = False
            if len(events) > 0 and events["t"][0] < previous_t:
                non_monotonic_ts += 1
                previous_t = events["t"][-1]
            non_monotonic_ts += numpy.count_nonzero(
                numpy.diff(events["t"].astype("<i8")) < 0
            )
            out_of_bounds_count += numpy.count_nonzero(
                numpy.logical_or(events["x"] >= file.width, events["y"] >= file.height)
            )
    except decode.RemainingBytesError as error:
        empty = False
        send_message(
            Error(path_id=file.path_id, message=f"{len(error.buffer)} extra bytes")
        )
    if empty:
        send_message(Error(path_id=file.path_id, message="no data found"))
    if out_of_bounds_count > 0:
        send_message(
            Error(
                path_id=file.path_id,
                message="{} event{} out of bounds".format(
                    out_of_bounds_count, "s" if out_of_bounds_count > 1 else ""
                ),
            )
        )
    if non_monotonic_ts > 0:
        send_message(
            Error(
                path_id=file.path_id,
                message="{} non-monotonic timestamp{}".format(
                    non_monotonic_ts, "s" if non_monotonic_ts > 1 else ""
                ),
            )
        )


def handle_imu(file: formats.ImuFile, send_message: formats.SendMessage):
    """Checks the invariants of an IMU file.

    Args:
        file (formats.ImuFile): The file to check.
        send_message (formats.SendMessage): Callback channel for errors.
    """
    non_monotonic_ts = 0
    empty = True
    try:
        previous_t = 0
        for imus in file.packets():
            empty = False
            if len(imus) > 0 and imus["t"][0] < previous_t:
                non_monotonic_ts += 1
                previous_t = imus["t"][-1]
            non_monotonic_ts += numpy.count_nonzero(
                numpy.diff(imus["t"].astype("<i8")) < 0
            )
    except decode.RemainingBytesError as error:
        send_message(
            Error(path_id=file.path_id, message=f"{len(error.buffer)} extra bytes")
        )
    if empty:
        send_message(Error(path_id=file.path_id, message="no data found"))
    if non_monotonic_ts > 0:
        send_message(
            Error(
                path_id=file.path_id,
                message="{} non-monotonic timestamp{}".format(
                    non_monotonic_ts, "s" if non_monotonic_ts > 1 else ""
                ),
            )
        )


def handle_other(file: path.File, send_message: formats.SendMessage):
    """Checks the invariants of an "other" file.

    This function simply consumes the file bytes.
    This guarantees that the file can be read and that the index contains the right hash and number of bytes.

    Args:
        file (formats.ImuFile): The file to check.
        send_message (formats.SendMessage): Callback channel for errors.
    """
    collections.deque(file.chunks(), maxlen=0)


def handle_directory(
    directory: path_directory.Directory, send_message: formats.SendMessage
):
    """Checks that system files are listed in the index, and vice-versa.

    Args:
        directory (path_directory.Directory): The directory to check.
        send_message (formats.SendMessage): Callback channel for errors.
    """
    index_data = json_index.load(directory.local_path / "-index.json")
    children = set(directory.local_path.iterdir())
    for file in itertools.chain(
        (
            formats.file_from_dict(data=data, parent=directory)
            for data in index_data["files"]
        ),
        (
            path.File.from_dict(data=data, parent=directory)
            for data in index_data["other_files"]
        ),
    ):
        local_path = utilities.path_with_suffix(
            file.local_path, file.best_compression.suffix
        )
        if not local_path in children:
            send_message(
                Error(
                    path_id=file.path_id,
                    message=f"{local_path} does not exist",
                )
            )
        children.remove(local_path)
        if not local_path.is_file():
            send_message(
                Error(
                    path_id=file.path_id,
                    message=f"{local_path} is not a file",
                )
            )
    for child_directory_name in index_data["directories"]:
        directory_path = directory.local_path / child_directory_name
        if not directory_path in children:
            send_message(
                Error(
                    path_id=directory.path_id,
                    message=f"{directory_path} does not exist",
                )
            )
        children.remove(directory_path)
        if not directory_path.is_dir():
            send_message(
                Error(
                    path_id=directory.path_id,
                    message=f"{directory_path} is not a directory",
                )
            )
    for child in children:
        if child.name != "-index.json":
            send_message(
                Error(
                    path_id=directory.path_id,
                    message=f"{child} is not listed in {directory.local_path / '-index.json'}",
                )
            )


def structure_recursive(path: pathlib.Path):
    """Rexucrively checks that the given path exists and that it has an UNDR structure (-index.json).

    Args:
        path (pathlib.Path): The local file path to check.

    Raises:
        RuntimeError: if the path is not a directory or does not contain a -index.json file.
    """
    if not path.exists():
        raise RuntimeError(f"{path} does not exist")
    if not path.is_dir():
        raise RuntimeError(f"{path} is not a directory")
    index_path = path / "-index.json"
    if not index_path.exists():
        raise RuntimeError(f"{index_path} does not exist")
    if not index_path.is_file():
        raise RuntimeError(f"{index_path} is not a file")
    index_data = json_index.load(index_path)
    for child_directory_name in index_data["directories"]:
        structure_recursive(path / child_directory_name)


def format_index_recursive(
    path: pathlib.Path, handle_path: typing.Callable[[pathlib.Path], None]
):
    """Validates the given index and formats its content.

    Args:
        path (pathlib.Path): Path of the index's parent directory.
        handle_path (typing.Callable[[pathlib.Path], None]): Called if the index was reformatted.
    """
    index_path = path / "-index.json"
    with open(index_path, "rb") as index_file:
        index_content = index_file.read()
    index_data = json.loads(index_content)
    json_index.schema.validate(index_data)
    new_index_content = f"{json.dumps(index_data, sort_keys=True, indent=4)}\n".encode()
    if index_content != new_index_content:
        handle_path(index_path)
        with open(index_path, "wb") as index_file:
            index_file.write(new_index_content)
    for child_directory_name in index_data["directories"]:
        format_index_recursive(
            path=path / child_directory_name, handle_path=handle_path
        )


class CheckFile(json_index_tasks.ProcessFile):
    """Uses a switch to process files and wraps messages into :py:class:`undr.check.Error`.

    Args:
        file (path.File): The file to process.
        switch (formats.Switch): Switch that maps file types to actions.
    """

    def __init__(self, file: path.File, switch: formats.Switch):
        super().__init__(file=file)
        self.switch = switch

    def run(self, session: requests.Session, manager: task.Manager):
        self.file.attach_session(session)
        self.file.attach_manager(manager)
        self.switch.handle_file(
            file=self.file,
            send_message=lambda message: manager.send_message(
                Error(path_id=self.file.path_id, message=message)
            ),
        )


class CheckLocalDirectoryRecursive(task.Task):
    """Dispatches check actions to validate the invariants of local files.

    This can be used to make sure that a dataset has been properly downloaded or to check files before uploading a dataset to a server.

    Args:
        path_root (pathlib.Path): The root path used to generate local file paths.
        path_id (pathlib.PurePosixPath): The path ID of the directory that will be scanned recursively.
        priority (int): Priority of this task and all recursively created tasks (tasks with lower priorities are scheduled first).
    """

    def __init__(
        self,
        path_root: pathlib.Path,
        path_id: pathlib.PurePosixPath,
        priority: int,
    ):
        self.path_root = path_root
        self.path_id = path_id
        self.priority = priority

    def run(self, session: requests.Session, manager: task.Manager):
        directory = path_directory.Directory(
            path_root=self.path_root,
            path_id=self.path_id,
            own_doi=None,
            metadata={},
            server=remote.NullServer(),
            doi_and_metadata_loaded=False,
        )
        handle_directory(
            directory=directory,
            send_message=lambda message: manager.send_message(
                Error(path_id=directory.path_id, message=message)
            ),
        )
        index_data = json_index.load(directory.local_path / "-index.json")
        for file in itertools.chain(
            (
                formats.file_from_dict(data=data, parent=directory)
                for data in index_data["files"]
            ),
            (
                path.File.from_dict(data=data, parent=directory)
                for data in index_data["other_files"]
            ),
        ):
            manager.schedule(
                CheckFile(
                    file=file,
                    switch=formats.Switch(
                        handle_aps=handle_aps,
                        handle_dvs=handle_dvs,
                        handle_imu=handle_imu,
                        handle_other=handle_other,
                    ),
                ),
                priority=self.priority,
            )
        for child_directory_name in index_data["directories"]:
            manager.schedule(
                CheckLocalDirectoryRecursive(
                    path_root=self.path_root,
                    path_id=self.path_id / child_directory_name,
                    priority=self.priority,
                ),
                priority=self.priority,
            )



from . import bibtex as bibtex
from . import check as check
from . import configuration as configuration
from . import constants as constants
from . import decode as decode
from . import display as display
from . import formats as formats
from . import install_mode as install_mode
from . import json_index as json_index
from . import json_index_tasks as json_index_tasks
from . import path as path
from . import path_directory as path_directory
from . import raw as raw
from . import remote as remote
from . import task as task
from . import utilities as utilities
from .configuration import Configuration as Configuration
from .configuration import IndexesStatuses as IndexesStatuses
from .configuration import IndexStatus as IndexStatus
from .configuration import configuration_from_path as configuration_from_path
from .decode import RemainingBytesError as RemainingBytesError
from .formats import ApsFile as ApsFile
from .formats import DvsFile as DvsFile
from .formats import ImuFile as ImuFile
from .formats import SendMessage as SendMessage
from .formats import Switch as Switch
from .install_mode import Mode as Mode
from .json_index_tasks import DirectoryScanned as DirectoryScanned
from .json_index_tasks import Index as Index
from .json_index_tasks import ProcessFile as ProcessFile
from .json_index_tasks import ProcessFilesRecursive as ProcessFilesRecursive
from .json_index_tasks import Selector as Selector
from .path import File as File
from .path_directory import Directory as Directory
from .persist import ReadOnlyStore as ReadOnlyStore
from .persist import Store as Store
from .simple import default_datasets as default_datasets
from .simple import install as install
from .task import Manager as Manager
from .task import ProcessManager as ProcessManager
from .task import Task as Task
from .task import WorkerException as WorkerException
from .version import __version__ as __version__



"""Installation strategies."""

from __future__ import annotations

import enum


class Mode(enum.Enum):
    """Download strategy for a dataset."""

    DISABLED = "disabled"
    """The dataset is ignored by all actions.
    """

    REMOTE = "remote"
    """Only download the dataset's index files.

    UNDR can process the dataset files as if they were locally available, by streaming them from the server. This option is particularly useful for large datasets that do not fit on the hard drive but it requires a fast internet connection since files are re-downloaded every time.
    """

    LOCAL = "local"
    """Download all the dataset files locally but do not decompress them.

    Most datasets are stored as Brotli archives (https://github.com/google/brotli/). UNDR stream-decompresses files before processing, making this option a good trade-off between disk usage and processing speed.
    """

    RAW = "raw"
    """Downloads all the dataset files locally and decompresses them.

    Decompressed files use a relatively inefficient plain binary file format. This option requires vast amounts of disk space (3 to 5 times as much as the Brotli archives). However, the plain binary format facilitates processing from languages such as Matlab or C++.
    """



"""High-level functions to install datasets without using a configuration file.

The functions in this module can be very convenient for one-off downloads and simple projects.
Configuration files (undr.toml) are recommended for more complex projects.
"""

from __future__ import annotations

import functools
import multiprocessing
import pathlib
import pkgutil
import typing

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

from . import configuration, constants, install_mode


@functools.lru_cache(maxsize=None)
def name_to_url() -> dict[str, str]:
    """Generates a dictionary of the default datasets' names and URLs.

    The first call to this function parses the configuration file bundled with UNDR and caches the result.
    Subsequent calls return the cached value immediately.
    Hence, modifying the returned dictionary also modifies the returned value of all subsequent calls, until the Python interpreter restarts.
    Users who plan to modify the returned value may want to call ``name_to_url().copy()`` to avoid this problem.

    Returns:
        dict[str, str]: Dictionary whose keys are dataset names and whose values are matching dataset URLs.
    """
    undr_default_bytes = pkgutil.get_data("undr", "undr_default.toml")
    assert undr_default_bytes is not None
    return {
        dataset["name"]: dataset["url"]
        for dataset in tomllib.loads(undr_default_bytes.decode())["datasets"]
    }


def default_datasets() -> list[str]:
    """Generates a list of the default datasets' names.

    This function calls :py:func:`name_to_url` and has the same caveats regarding caching.

    Returns:
        list[str]: The names of the default datasets.
    """
    return list(name_to_url().keys())


def install(
    name: str,
    url: typing.Optional[str] = None,
    timeout: float = constants.DEFAULT_TIMEOUT,
    mode: typing.Union[str, install_mode.Mode] = install_mode.Mode.LOCAL,
    directory: typing.Union[str, pathlib.Path] = "datasets",
    show_display: bool = True,
    workers: int = multiprocessing.cpu_count() * 2,
    force: bool = False,
    log_directory: typing.Optional[pathlib.Path] = None,
):
    """Downloads (and optionally decompresses) a dataset.

    See :py:class:`undr.install_mode.Mode` for details on the different installation strategies.

    Args:
        name (str): Name of the dataset to install. Unless url is provided, it must be one of the keys returned by :py:func:`name_to_url`.
        url (typing.Optional[str], optional): URL of the dataset. Defaults to None.
        timeout (float, optional): Request timeout in seconds. Defaults to :py:attr:`undr.constants.DEFAULT_TIMEOUT`.
        mode (typing.Union[str, install_mode.Mode], optional): Installation strategy. Defaults to :py:attr:`undr.install_mode.Mode.LOCAL`.
        directory (typing.Union[str, pathlib.Path], optional): Path of the local directory to store datasets. Defaults to "datasets".
        show_display (bool, optional): Whether to show a progress bar. Defaults to True.
        workers (int, optional): Number of parallel workers (threads). Defaults to twice :py:func:`multiprocessing.cpu_count`
        force (bool, optional): Whether to re-download files even if they are already present locally. Defaults to False.
        log_directory (typing.Optional[pathlib.Path], optional): Directory to store log files. Logs are not generated if this is None. Defaults to None.
    """
    if url is None:
        url = name_to_url()[name]
    if isinstance(mode, str):
        mode = install_mode.Mode(mode)
    if isinstance(directory, str):
        directory = pathlib.Path(directory)
    directory = directory.resolve()
    directory.mkdir(exist_ok=True, parents=True)
    configuration.Configuration(
        directory=directory,
        name_to_dataset_settings={
            name: configuration.DatasetSettings(
                name=name,
                url=url,
                mode=mode,
                timeout=timeout,
            )
        },
    ).install(
        show_display=show_display,
        workers=workers,
        force=force,
        log_directory=log_directory,
    )



import pathlib
import sys

import setuptools

dirname = pathlib.Path(__file__).resolve().parent

if (
    not "-h" in sys.argv
    and not "--help" in sys.argv
    and (dirname.parent / "specification" / "undr_schema.json").is_file()
):
    import shutil

    shutil.rmtree(dirname / "undr", ignore_errors=True)
    shutil.copytree(dirname / "source", dirname / "undr")
    shutil.copyfile(
        dirname.parent / "specification" / "undr_schema.json",
        dirname / "undr" / "undr_schema.json",
    )
    shutil.copyfile(
        dirname.parent / "specification" / "-index_schema.json",
        dirname / "undr" / "-index_schema.json",
    )
    shutil.copyfile(
        dirname.parent / "undr_default.toml", dirname / "undr" / "undr_default.toml"
    )
    shutil.copyfile(dirname.parent / "README.md", dirname / "long_description.md")

with open(dirname / "long_description.md") as long_description_file:
    long_description = long_description_file.read()

exec(open(dirname / "undr" / "version.py").read())  # defines __version__

setuptools.setup(
    name="undr",
    version=__version__,  # type: ignore
    url="https://github.com/neuromorphicsystems/undr",
    author="Alexandre Marcireau",
    author_email="alexandre.marcireau@gmail.com",
    description="Download and process Neuromorphic datasets",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "brotli>=1.0,<2",
        "jsonschema-rs>=0.17",
        "numpy>=1.22,<2",
        "requests>=2.31,<3",
        "tomli>=2.0.0; python_version<\"3.11\"",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    packages=["undr"],
    package_data={"": ["*.json", "*.toml"]},
)



"""Local or remote directory."""

from __future__ import annotations

import dataclasses
import functools
import itertools
import typing

from . import constants, formats, json_index, path


@dataclasses.dataclass(frozen=True)
class Directory(path.Path):
    doi_and_metadata_loaded: bool = dataclasses.field(
        repr=False, compare=False, hash=False
    )

    def __getattribute__(self, __name: str):
        if not object.__getattribute__(self, "doi_and_metadata_loaded") and (
            __name == "metadata" or __name == "own_doi"
        ):
            index_data = json_index.load(self.local_path / "-index.json")
            if "doi" in index_data:
                object.__setattr__(self, "own_doi", index_data["doi"])
            if "metadata" in index_data:
                object.__setattr__(self, "metadata", index_data["metadata"])
            object.__setattr__(self, "doi_and_metadata_loaded", True)
        return super().__getattribute__(__name)

    @functools.lru_cache(maxsize=constants.LRU_CACHE_MAXSIZE)
    def __truediv__(self, other: str) -> path.Path:  # type: ignore
        index_data = json_index.load(self.local_path / "-index.json")
        for child_directory_name in index_data["directories"]:
            if other == child_directory_name:
                return Directory(
                    path_root=self.path_root,
                    path_id=self.path_id / child_directory_name,
                    own_doi=None,
                    metadata={},
                    server=self.server,
                    doi_and_metadata_loaded=False,
                )
        for file_data in index_data["files"]:
            if other == file_data["name"]:
                return formats.file_from_dict(data=file_data, parent=self)
        for file_data in index_data["other_files"]:
            if other == file_data["name"]:
                return path.File.from_dict(data=file_data, parent=self)
        raise Exception(f'"{other}" not found in "{self.local_path / "-index.json"}"')

    def iter(self, recursive: bool = False) -> typing.Iterable[path.Path]:
        index_data = json_index.load(self.local_path / "-index.json")
        if "doi" in index_data:
            self.__dict__["own_doi"] = index_data["doi"]
        if "metadata" in index_data:
            self.__dict__["metadata"] = index_data["metadata"]
        object.__setattr__(self, "doi_and_metadata_loaded", True)
        yield self
        for file in itertools.chain(
            (
                formats.file_from_dict(data=data, parent=self)
                for data in index_data["files"]
            ),
            (
                path.File.from_dict(data=data, parent=self)
                for data in index_data["other_files"]
            ),
        ):
            yield file
        for child_directory_name in index_data["directories"]:
            child_directory = Directory(
                path_root=self.path_root,
                path_id=self.path_id / child_directory_name,
                own_doi=None,
                metadata={},
                server=self.server,
                doi_and_metadata_loaded=False,
            )
            if recursive:
                yield from child_directory.iter(recursive=True)
            else:
                yield child_directory
